from builtins import str
from builtins import object
class Chunk(object):

   def __str__(self):
        return str(self)


class TagChunk(Chunk):

    def __init__(self, tag, label=None):
        self.tag = tag
        self.label = label

    def __unicode__(self):
        if self.label is None:
            return self.tag
        else:
            return self.label + ":" + self.tag

class TextChunk(Chunk):

    def __init__(self, text):
        self.text = text

    def __unicode__(self):
        return "'" + self.text + "'"




#
# [The "BSD license"]
#  Copyright (c) 2012 Terence Parr
#  Copyright (c) 2012 Sam Harwell
#  Copyright (c) 2014 Eric Vergnaud
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#  3. The name of the author may not be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
#  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
#  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
#  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
#  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
#  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
#  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

#  An ATN transition between any two ATN states.  Subclasses define
#  atom, set, epsilon, action, predicate, rule transitions.
#
#  <p>This is a one way link.  It emanates from a state (usually via a list of
#  transitions) and has a target state.</p>
#
#  <p>Since we never have to change the ATN transitions once we construct it,
#  we can fix these transitions as specific classes. The DFA transitions
#  on the other hand need to update the labels as it adds transitions to
#  the states. We'll use the term Edge for the DFA to distinguish them from
#  ATN transitions.</p>
#
from future import standard_library
standard_library.install_aliases()
from builtins import str
from builtins import chr
from builtins import object
from builtins import str
from custom_antlr4.IntervalSet import IntervalSet, Interval
from custom_antlr4.Token import Token
from custom_antlr4.atn.SemanticContext import Predicate, PrecedencePredicate

class Transition (object):
    # constants for serialization
    EPSILON			= 1
    RANGE			= 2
    RULE			= 3
    PREDICATE		= 4 # e.g., {isType(input.LT(1))}?
    ATOM			= 5
    ACTION			= 6
    SET				= 7 # ~(A|B) or ~atom, wildcard, which convert to next 2
    NOT_SET			= 8
    WILDCARD		= 9
    PRECEDENCE		= 10

    serializationNames = [
            u"INVALID",
            u"EPSILON",
            u"RANGE",
            u"RULE",
            u"PREDICATE",
            u"ATOM",
            u"ACTION",
            u"SET",
            u"NOT_SET",
            u"WILDCARD",
            u"PRECEDENCE"
        ]

    serializationTypes = dict()

    def __init__(self, target):
        # The target of this transition.
        if target is None:
            raise Exception("target cannot be null.")
        self.target = target
        # Are we epsilon, action, sempred?
        self.isEpsilon = False
        self.label = None




# TODO: make all transitions sets? no, should remove set edges
class AtomTransition(Transition):

    def __init__(self, target, label):
        super(AtomTransition, self).__init__(target)
        self.label_ = label # The token type or character value; or, signifies special label.
        self.label = self.makeLabel()
        self.serializationType = self.ATOM

    def makeLabel(self):
        s = IntervalSet()
        s.addOne(self.label_)
        return s

    def matches( self, symbol, minVocabSymbol,  maxVocabSymbol):
        return self.label_ == symbol

    def __unicode__(self):
        return str(self.label_)

class RuleTransition(Transition):

    def __init__(self, ruleStart, ruleIndex, precedence, followState):
        super(RuleTransition, self).__init__(ruleStart)
        self.ruleIndex = ruleIndex # ptr to the rule definition object for this rule ref
        self.precedence = precedence
        self.followState = followState # what node to begin computations following ref to rule
        self.serializationType = self.RULE
        self.isEpsilon = True

    def matches( self, symbol, minVocabSymbol,  maxVocabSymbol):
        return False


class EpsilonTransition(Transition):

    def __init__(self, target):
        super(EpsilonTransition, self).__init__(target)
        self.serializationType = self.EPSILON
        self.isEpsilon = True

    def matches( self, symbol, minVocabSymbol,  maxVocabSymbol):
        return False

    def __unicode__(self):
        return "epsilon"

class RangeTransition(Transition):

    def __init__(self, target, start, stop):
        super(RangeTransition, self).__init__(target)
        self.serializationType = self.RANGE
        self.start = start
        self.stop = stop
        self.label = self.makeLabel()

    def makeLabel(self):
        s = IntervalSet()
        s.addRange(Interval(self.start, self.stop + 1))
        return s

    def matches( self, symbol, minVocabSymbol,  maxVocabSymbol):
        return symbol >= self.start and symbol <= self.stop

    def __unicode__(self):
        return "'" + chr(self.start) + "'..'" + chr(self.stop) + "'"

class AbstractPredicateTransition(Transition):

    def __init__(self, target):
        super(AbstractPredicateTransition, self).__init__(target)


class PredicateTransition(AbstractPredicateTransition):

    def __init__(self, target, ruleIndex, predIndex, isCtxDependent):
        super(PredicateTransition, self).__init__(target)
        self.serializationType = self.PREDICATE
        self.ruleIndex = ruleIndex
        self.predIndex = predIndex
        self.isCtxDependent = isCtxDependent # e.g., $i ref in pred
        self.isEpsilon = True

    def matches( self, symbol, minVocabSymbol,  maxVocabSymbol):
        return False

    def getPredicate(self):
        return Predicate(self.ruleIndex, self.predIndex, self.isCtxDependent)

    def __unicode__(self):
        return u"pred_" + str(self.ruleIndex) + u":" + str(self.predIndex)

class ActionTransition(Transition):

    def __init__(self, target, ruleIndex, actionIndex=-1, isCtxDependent=False):
        super(ActionTransition, self).__init__(target)
        self.serializationType = self.ACTION
        self.ruleIndex = ruleIndex
        self.actionIndex = actionIndex
        self.isCtxDependent = isCtxDependent # e.g., $i ref in pred
        self.isEpsilon = True

    def matches( self, symbol, minVocabSymbol,  maxVocabSymbol):
        return False

    def __unicode__(self):
        return u"action_" + str(self.ruleIndex) + u":" + str(self.actionIndex)

# A transition containing a set of values.
class SetTransition(Transition):

    def __init__(self, target, set):
        super(SetTransition, self).__init__(target)
        self.serializationType = self.SET
        if set is not None:
            self.label = set
        else:
            self.label = IntervalSet()
            self.label.addRange(Interval(Token.INVALID_TYPE, Token.INVALID_TYPE + 1))

    def matches( self, symbol, minVocabSymbol,  maxVocabSymbol):
        return symbol in self.label

    def __unicode__(self):
        return str(self.label)

class NotSetTransition(SetTransition):

    def __init__(self, target, set):
        super(NotSetTransition, self).__init__(target, set)
        self.serializationType = self.NOT_SET

    def matches( self, symbol, minVocabSymbol,  maxVocabSymbol):
        return symbol >= minVocabSymbol \
            and symbol <= maxVocabSymbol \
            and not super(type(self), self).matches(symbol, minVocabSymbol, maxVocabSymbol)

    def __unicode__(self):
        return u'~' + super(type(self), self).__unicode__()


class WildcardTransition(Transition):

    def __init__(self, target):
        super(WildcardTransition, self).__init__(target)
        self.serializationType = self.WILDCARD

    def matches( self, symbol, minVocabSymbol,  maxVocabSymbol):
        return symbol >= minVocabSymbol and symbol <= maxVocabSymbol

    def __unicode__(self):
        return u"."


class PrecedencePredicateTransition(AbstractPredicateTransition):

    def __init__(self, target, precedence):
        super(PrecedencePredicateTransition, self).__init__(target)
        self.serializationType = self.PRECEDENCE
        self.precedence = precedence
        self.isEpsilon = True

    def matches( self, symbol, minVocabSymbol,  maxVocabSymbol):
        return False


    def getPredicate(self):
        return PrecedencePredicate(self.precedence)

    def __unicode__(self):
        return self.precedence + " >= _p"


Transition.serializationTypes = {
             EpsilonTransition: Transition.EPSILON,
             RangeTransition: Transition.RANGE,
             RuleTransition: Transition.RULE,
             PredicateTransition: Transition.PREDICATE,
             AtomTransition: Transition.ATOM,
             ActionTransition: Transition.ACTION,
             SetTransition: Transition.SET,
             NotSetTransition: Transition.NOT_SET,
             WildcardTransition: Transition.WILDCARD,
             PrecedencePredicateTransition: Transition.PRECEDENCE
         }




# Generated from java-escape by ANTLR 4.4
from custom_antlr4 import *


# This class defines a complete listener for a parse tree produced by JavaParser.
class JavaListener(ParseTreeListener):
    # Enter a parse tree produced by JavaParser#memberDeclaration.
    def enterMemberDeclaration(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#memberDeclaration.
    def exitMemberDeclaration(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#defaultValue.
    def enterDefaultValue(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#defaultValue.
    def exitDefaultValue(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#annotationTypeElementDeclaration.
    def enterAnnotationTypeElementDeclaration(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#annotationTypeElementDeclaration.
    def exitAnnotationTypeElementDeclaration(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#annotationTypeBody.
    def enterAnnotationTypeBody(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#annotationTypeBody.
    def exitAnnotationTypeBody(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#lambdaBody.
    def enterLambdaBody(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#lambdaBody.
    def exitLambdaBody(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#genericInterfaceMethodDeclaration.
    def enterGenericInterfaceMethodDeclaration(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#genericInterfaceMethodDeclaration.
    def exitGenericInterfaceMethodDeclaration(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#classBodyDeclaration.
    def enterClassBodyDeclaration(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#classBodyDeclaration.
    def exitClassBodyDeclaration(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#block.
    def enterBlock(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#block.
    def exitBlock(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#enumBodyDeclarations.
    def enterEnumBodyDeclarations(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#enumBodyDeclarations.
    def exitEnumBodyDeclarations(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#forUpdate.
    def enterForUpdate(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#forUpdate.
    def exitForUpdate(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#enhancedForControl.
    def enterEnhancedForControl(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#enhancedForControl.
    def exitEnhancedForControl(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#annotationConstantRest.
    def enterAnnotationConstantRest(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#annotationConstantRest.
    def exitAnnotationConstantRest(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#explicitGenericInvocation.
    def enterExplicitGenericInvocation(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#explicitGenericInvocation.
    def exitExplicitGenericInvocation(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#nonWildcardTypeArgumentsOrDiamond.
    def enterNonWildcardTypeArgumentsOrDiamond(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#nonWildcardTypeArgumentsOrDiamond.
    def exitNonWildcardTypeArgumentsOrDiamond(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#expressionList.
    def enterExpressionList(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#expressionList.
    def exitExpressionList(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#annotationTypeElementRest.
    def enterAnnotationTypeElementRest(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#annotationTypeElementRest.
    def exitAnnotationTypeElementRest(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#classOrInterfaceType.
    def enterClassOrInterfaceType(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#classOrInterfaceType.
    def exitClassOrInterfaceType(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#jType.
    def enterJType(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#jType.
    def exitJType(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#typeBound.
    def enterTypeBound(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#typeBound.
    def exitTypeBound(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#variableDeclaratorId.
    def enterVariableDeclaratorId(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#variableDeclaratorId.
    def exitVariableDeclaratorId(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#primary.
    def enterPrimary(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#primary.
    def exitPrimary(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#classCreatorRest.
    def enterClassCreatorRest(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#classCreatorRest.
    def exitClassCreatorRest(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#interfaceBodyDeclaration.
    def enterInterfaceBodyDeclaration(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#interfaceBodyDeclaration.
    def exitInterfaceBodyDeclaration(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#typeArguments.
    def enterTypeArguments(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#typeArguments.
    def exitTypeArguments(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#annotationName.
    def enterAnnotationName(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#annotationName.
    def exitAnnotationName(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#inferredFormalParameterList.
    def enterInferredFormalParameterList(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#inferredFormalParameterList.
    def exitInferredFormalParameterList(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#finallyBlock.
    def enterFinallyBlock(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#finallyBlock.
    def exitFinallyBlock(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#typeParameters.
    def enterTypeParameters(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#typeParameters.
    def exitTypeParameters(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#lastFormalParameter.
    def enterLastFormalParameter(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#lastFormalParameter.
    def exitLastFormalParameter(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#constructorBody.
    def enterConstructorBody(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#constructorBody.
    def exitConstructorBody(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#literal.
    def enterLiteral(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#literal.
    def exitLiteral(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#annotationMethodOrConstantRest.
    def enterAnnotationMethodOrConstantRest(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#annotationMethodOrConstantRest.
    def exitAnnotationMethodOrConstantRest(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#catchClause.
    def enterCatchClause(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#catchClause.
    def exitCatchClause(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#variableDeclarator.
    def enterVariableDeclarator(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#variableDeclarator.
    def exitVariableDeclarator(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#typeList.
    def enterTypeList(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#typeList.
    def exitTypeList(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#enumConstants.
    def enterEnumConstants(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#enumConstants.
    def exitEnumConstants(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#classBody.
    def enterClassBody(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#classBody.
    def exitClassBody(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#createdName.
    def enterCreatedName(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#createdName.
    def exitCreatedName(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#enumDeclaration.
    def enterEnumDeclaration(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#enumDeclaration.
    def exitEnumDeclaration(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#formalParameter.
    def enterFormalParameter(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#formalParameter.
    def exitFormalParameter(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#parExpression.
    def enterParExpression(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#parExpression.
    def exitParExpression(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#annotation.
    def enterAnnotation(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#annotation.
    def exitAnnotation(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#variableInitializer.
    def enterVariableInitializer(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#variableInitializer.
    def exitVariableInitializer(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#elementValueArrayInitializer.
    def enterElementValueArrayInitializer(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#elementValueArrayInitializer.
    def exitElementValueArrayInitializer(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#creator.
    def enterCreator(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#creator.
    def exitCreator(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#arrayCreatorRest.
    def enterArrayCreatorRest(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#arrayCreatorRest.
    def exitArrayCreatorRest(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#expression.
    def enterExpression(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#expression.
    def exitExpression(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#constantExpression.
    def enterConstantExpression(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#constantExpression.
    def exitConstantExpression(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#qualifiedNameList.
    def enterQualifiedNameList(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#qualifiedNameList.
    def exitQualifiedNameList(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#lambdaParameters.
    def enterLambdaParameters(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#lambdaParameters.
    def exitLambdaParameters(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#constructorDeclaration.
    def enterConstructorDeclaration(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#constructorDeclaration.
    def exitConstructorDeclaration(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#forControl.
    def enterForControl(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#forControl.
    def exitForControl(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#superSuffix.
    def enterSuperSuffix(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#superSuffix.
    def exitSuperSuffix(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#variableDeclarators.
    def enterVariableDeclarators(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#variableDeclarators.
    def exitVariableDeclarators(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#catchType.
    def enterCatchType(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#catchType.
    def exitCatchType(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#classOrInterfaceModifier.
    def enterClassOrInterfaceModifier(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#classOrInterfaceModifier.
    def exitClassOrInterfaceModifier(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#enumConstantName.
    def enterEnumConstantName(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#enumConstantName.
    def exitEnumConstantName(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#modifier.
    def enterModifier(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#modifier.
    def exitModifier(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#innerCreator.
    def enterInnerCreator(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#innerCreator.
    def exitInnerCreator(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#explicitGenericInvocationSuffix.
    def enterExplicitGenericInvocationSuffix(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#explicitGenericInvocationSuffix.
    def exitExplicitGenericInvocationSuffix(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#variableModifier.
    def enterVariableModifier(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#variableModifier.
    def exitVariableModifier(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#elementValuePair.
    def enterElementValuePair(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#elementValuePair.
    def exitElementValuePair(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#arrayInitializer.
    def enterArrayInitializer(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#arrayInitializer.
    def exitArrayInitializer(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#elementValue.
    def enterElementValue(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#elementValue.
    def exitElementValue(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#lambdaExpression.
    def enterLambdaExpression(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#lambdaExpression.
    def exitLambdaExpression(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#constDeclaration.
    def enterConstDeclaration(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#constDeclaration.
    def exitConstDeclaration(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#resource.
    def enterResource(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#resource.
    def exitResource(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#qualifiedName.
    def enterQualifiedName(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#qualifiedName.
    def exitQualifiedName(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#resourceSpecification.
    def enterResourceSpecification(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#resourceSpecification.
    def exitResourceSpecification(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#formalParameterList.
    def enterFormalParameterList(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#formalParameterList.
    def exitFormalParameterList(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#annotationTypeDeclaration.
    def enterAnnotationTypeDeclaration(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#annotationTypeDeclaration.
    def exitAnnotationTypeDeclaration(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#compilationUnit.
    def enterCompilationUnit(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#compilationUnit.
    def exitCompilationUnit(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#annotationMethodRest.
    def enterAnnotationMethodRest(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#annotationMethodRest.
    def exitAnnotationMethodRest(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#switchBlockStatementGroup.
    def enterSwitchBlockStatementGroup(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#switchBlockStatementGroup.
    def exitSwitchBlockStatementGroup(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#typeParameter.
    def enterTypeParameter(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#typeParameter.
    def exitTypeParameter(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#interfaceBody.
    def enterInterfaceBody(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#interfaceBody.
    def exitInterfaceBody(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#methodDeclaration.
    def enterMethodDeclaration(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#methodDeclaration.
    def exitMethodDeclaration(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#methodBody.
    def enterMethodBody(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#methodBody.
    def exitMethodBody(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#typeArgument.
    def enterTypeArgument(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#typeArgument.
    def exitTypeArgument(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#typeDeclaration.
    def enterTypeDeclaration(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#typeDeclaration.
    def exitTypeDeclaration(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#genericConstructorDeclaration.
    def enterGenericConstructorDeclaration(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#genericConstructorDeclaration.
    def exitGenericConstructorDeclaration(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#classDeclaration.
    def enterClassDeclaration(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#classDeclaration.
    def exitClassDeclaration(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#enumConstant.
    def enterEnumConstant(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#enumConstant.
    def exitEnumConstant(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#statement.
    def enterStatement(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#statement.
    def exitStatement(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#importDeclaration.
    def enterImportDeclaration(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#importDeclaration.
    def exitImportDeclaration(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#primitiveType.
    def enterPrimitiveType(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#primitiveType.
    def exitPrimitiveType(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#interfaceDeclaration.
    def enterInterfaceDeclaration(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#interfaceDeclaration.
    def exitInterfaceDeclaration(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#localVariableDeclarationStatement.
    def enterLocalVariableDeclarationStatement(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#localVariableDeclarationStatement.
    def exitLocalVariableDeclarationStatement(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#blockStatement.
    def enterBlockStatement(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#blockStatement.
    def exitBlockStatement(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#fieldDeclaration.
    def enterFieldDeclaration(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#fieldDeclaration.
    def exitFieldDeclaration(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#constantDeclarator.
    def enterConstantDeclarator(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#constantDeclarator.
    def exitConstantDeclarator(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#resources.
    def enterResources(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#resources.
    def exitResources(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#statementExpression.
    def enterStatementExpression(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#statementExpression.
    def exitStatementExpression(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#interfaceMethodDeclaration.
    def enterInterfaceMethodDeclaration(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#interfaceMethodDeclaration.
    def exitInterfaceMethodDeclaration(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#packageDeclaration.
    def enterPackageDeclaration(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#packageDeclaration.
    def exitPackageDeclaration(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#elementValuePairs.
    def enterElementValuePairs(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#elementValuePairs.
    def exitElementValuePairs(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#localVariableDeclaration.
    def enterLocalVariableDeclaration(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#localVariableDeclaration.
    def exitLocalVariableDeclaration(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#nonWildcardTypeArguments.
    def enterNonWildcardTypeArguments(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#nonWildcardTypeArguments.
    def exitNonWildcardTypeArguments(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#interfaceMemberDeclaration.
    def enterInterfaceMemberDeclaration(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#interfaceMemberDeclaration.
    def exitInterfaceMemberDeclaration(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#switchLabel.
    def enterSwitchLabel(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#switchLabel.
    def exitSwitchLabel(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#forInit.
    def enterForInit(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#forInit.
    def exitForInit(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#formalParameters.
    def enterFormalParameters(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#formalParameters.
    def exitFormalParameters(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#arguments.
    def enterArguments(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#arguments.
    def exitArguments(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#genericMethodDeclaration.
    def enterGenericMethodDeclaration(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#genericMethodDeclaration.
    def exitGenericMethodDeclaration(self, ctx):
        pass

    # Enter a parse tree produced by JavaParser#typeArgumentsOrDiamond.
    def enterTypeArgumentsOrDiamond(self, ctx):
        pass

    # Exit a parse tree produced by JavaParser#typeArgumentsOrDiamond.
    def exitTypeArgumentsOrDiamond(self, ctx):
        pass



from __future__ import print_function
from __future__ import division

"""
    __     _  __   __   __       ____                          _
   / /    (_)/ /_ / /_ / /___   / __ \ ____ _ _____ _      __ (_)____
  / /    / // __// __// // _ \ / / / // __ `// ___/| | /| / // // __ \
 / /___ / // /_ / /_ / //  __// /_/ // /_/ // /    | |/ |/ // // / / /
/_____//_/ \__/ \__//_/ \___//_____/ \__,_//_/     |__/|__//_//_/ /_/

Copyright (c) 2014-2020 Ali Parsai

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

Ali Parsai
www.parsai.net
ali@parsai.net

"""

# generic imports
import io
import sys
import platform
import os
from optparse import OptionParser
import shelve
import shutil
import datetime
import threading
import signal
import time
import subprocess

from builtins import str
from past.utils import old_div

### DEBUG ###


# def trace(frame, event, arg):
#     print("%s, %s:%d" % (event, frame.f_code.co_filename, frame.f_lineno))
#     return trace

# sys.settrace(trace)

#############


# LittleDarwin modules
from .JavaParse import JavaParse
from .JavaRead import JavaRead
from .JavaMutate import JavaMutate
from .ReportGenerator import ReportGenerator
from littledarwin import License


littleDarwinVersion = "0.5.1"


# Alternative to subprocess32

# this method uses threading backend to create a watchdog thread that kills the build system and any child processes
# after the timeout is passed.


def timeoutAlternative(commandString, workingDirectory, timeout, inputData=None):
    """

    :param commandString: command to run
    :param workingDirectory: the directory that the command is supposed to run in
    :param timeout: timeout in seconds
    :param inputData: the data that the run process may need. defaults to None.
    :return: returns kill status, process return code and the output of the system
    """

    killCheck = threading.Event()

    # this method is run in another thread when the timeout is expired to kill the process.
    def killProcess(pipe):

        assert isinstance(pipe, subprocess.Popen)

        # there is no support for os.killpg on windows, neither does it have SIGKILL.
        if platform.system() == "Windows":
            # this utility is not included in windows XP Home edition, however, there is no other alternative either.
            # therefore, don't run LittleDarwin on windows XP Home edition; he gets sad.
            subprocess.Popen("taskkill /F /T /PID %i" % pipe.pid, shell=True)
        else:
            # posix systems all support this call.
            # pipe.terminate()
            try:
                os.killpg(os.getpgid(pipe.pid), signal.SIGTERM)
            except:
                os.kill(pipe.pid, signal.SIGTERM)

        # we just killed the process. let everyone know.
        killCheck.set()

    # timeout must be int, otherwise problems arise.
    assert isinstance(timeout, int)

    # starting the process with the given parameters.
    if platform.system() != "Windows":
        process = subprocess.Popen(commandString, bufsize=1, cwd=workingDirectory, stdin=subprocess.PIPE,
                                   stdout=subprocess.PIPE, stderr=subprocess.STDOUT, preexec_fn=os.setsid)
    else:  # in Windows
        process = subprocess.Popen(commandString, bufsize=1, cwd=workingDirectory, stdin=subprocess.PIPE,
                                   stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

    # passing the process and timeout references to threading's timer method, so that it kills the process
    # if timeout expires.
    timerWatchdog = threading.Timer(timeout, killProcess, args=[process])
    timerWatchdog.start()

    # do the stuff in the process.
    (stdout, stderr) = process.communicate(inputData)

    # if the process is done, no need to kill it.
    timerWatchdog.cancel()

    isKilled = killCheck.isSet()
    killCheck.clear()

    return isKilled, process.returncode, stdout


def main(argv):
    print("""
    __     _  __   __   __       ____                          _
   / /    (_)/ /_ / /_ / /___   / __ \ ____ _ _____ _      __ (_)____
  / /    / // __// __// // _ \ / / / // __ `// ___/| | /| / // // __ \\
 / /___ / // /_ / /_ / //  __// /_/ // /_/ // /    | |/ |/ // // / / /
/_____//_/ \__/ \__//_/ \___//_____/ \__,_//_/     |__/|__//_//_/ /_/

      _                     _                 ___
     /_|  /|/|  _/__/'     /_|   _ /   _ ' _ (_  _ _ _  _      _ /
    (  | /   |(//(///()/) (  |/)(/((/_) /_)  /  / (///)(-((/()/ /(
                                  /


    LittleDarwin version %s Copyright (C) 2014-2020 Ali Parsai

    LittleDarwin comes with ABSOLUTELY NO WARRANTY.
    This is free software, and you are welcome to redistribute it
    under certain conditions; run LittleDarwin --license for details.


    """ % littleDarwinVersion)

    # parsing input options
    optionParser = OptionParser(prog="LittleDarwin")

    optionParser.add_option("-m", "--mutate", action="store_true", dest="isMutationActive", default=False,
                            help="Activate the mutation phase.")
    optionParser.add_option("-b", "--build", action="store_true", dest="isBuildActive", default=False,
                            help="Activate the build phase.")
    optionParser.add_option("-v", "--verbose", action="store_true", dest="isVerboseActive", default=False,
                            help="Verbose output.")
    optionParser.add_option("-p", "--path", action="store", dest="sourcePath",
                            default=os.path.dirname(os.path.realpath(__file__)), help="Path to source files.")
    optionParser.add_option("-t", "--build-path", action="store", dest="buildPath",
                            default=os.path.dirname(os.path.realpath(__file__)),
                            help="Path to build system working directory.")
    optionParser.add_option("-c", "--build-command", action="store", dest="buildCommand", default="mvn,test",
                            help="Command to run the build system. If it includes more than a single argument, they should be seperated by comma. For example: mvn,install")
    optionParser.add_option("--test-path", action="store", dest="testPath",
                            default="***dummy***", help="path to test project build system working directory")
    optionParser.add_option("--test-command", action="store", dest="testCommand", default="***dummy***",
                            help="Command to run the test-suite. If it includes more than a single argument, they should be seperated by comma. For example: mvn,test")
    optionParser.add_option("--initial-build-command", action="store", dest="initialBuildCommand",
                            default="***dummy***", help="Command to run the initial build.")
    optionParser.add_option("--timeout", type="int", action="store", dest="timeout", default=60,
                            help="Timeout value for the build process.")
    optionParser.add_option("--cleanup", action="store", dest="cleanUp", default="***dummy***",
                            help="Commands to run after each build.")
    optionParser.add_option("--use-alternate-database", action="store", dest="alternateDb", default="***dummy***",
                            help="Path to alternative database.")
    optionParser.add_option("--license", action="store_true", dest="isLicenseActive", default=False,
                            help="Output the license and exit.")
    optionParser.add_option("--higher-order", type="int", action="store", dest="higherOrder", default=1,
                            help="Define order of mutation. Use -1 to dynamically adjust per class.")
    optionParser.add_option("--null-check", action="store_true", dest="isNullCheck", default=False,
                            help="Use null check mutation operators.")
    optionParser.add_option("--all", action="store_true", dest="isAll", default=False,
                            help="Use all mutation operators.")
    optionParser.add_option("--whitelist", action="store", dest="whitelist", default="***dummy***",
                            help="Analyze only included packages or files defined in this file (one package name or path to file per line).")
    optionParser.add_option("--blacklist", action="store", dest="blacklist", default="***dummy***",
                            help="Analyze everything except packages or files defined in this file (one package name or path to file per line).")

    (options, args) = optionParser.parse_args()

    if options.whitelist != "***dummy***" and options.blacklist != "***dummy***":
        print("You can either define a whitelist or a blacklist but not both.")
        sys.exit(0)

    filterList = None
    filterType = None
    if options.whitelist != "***dummy***" and os.path.isfile(options.whitelist):
        with io.open(options.whitelist, mode='r', errors='replace') as contentFile:
            filterList = contentFile.readlines()
            filterType = "whitelist"

    if options.blacklist != "***dummy***" and os.path.isfile(options.blacklist):
        with io.open(options.blacklist, mode='r', errors='replace') as contentFile:
            filterList = contentFile.readlines()
            filterType = "blacklist"

    if filterList is not None:
        filterList = [_f for _f in filterList if _f]

    if options.isLicenseActive:
        License.outputLicense()
        sys.exit(0)

    if options.higherOrder <= 1 and options.higherOrder != -1:
        higherOrder = 1
    else:
        higherOrder = options.higherOrder

    # there is an upside in not running two phases together. we may include the ability to edit some mutants later.
    if options.isBuildActive and options.isMutationActive:
        print("it is strongly recommended to do the analysis in two different phases.\n\n")

    # *****************************************************************************************************************
    # ---------------------------------------- mutant generation phase ------------------------------------------------
    # *****************************************************************************************************************

    if options.isMutationActive:
        assert options.isVerboseActive is not None
        # creating our module objects.
        javaRead = JavaRead(options.isVerboseActive)
        javaParse = JavaParse(options.isVerboseActive)
        javaMutate = JavaMutate(javaParse, options.isVerboseActive)

        totalMutantCount = 0

        try:
            assert os.path.isdir(options.sourcePath)
        except AssertionError as exception:
            print("source path must be a directory.")
            sys.exit(1)

        # getting the list of files.
        javaRead.listFiles(targetPath=os.path.abspath(options.sourcePath), buildPath=os.path.abspath(options.buildPath),
                           filterType=filterType, filterList=filterList)
        fileCounter = 0
        fileCount = len(javaRead.fileList)

        # creating a database for generated mutants. the format of this database is different on different platforms,
        # so it cannot be simply copied from a platform to another.
        databasePath = os.path.join(javaRead.targetDirectory, "mutationdatabase")

        print("source dir: ", javaRead.sourceDirectory)
        print("target dir: ", javaRead.targetDirectory)
        print("creating mutation database: ", databasePath)

        mutationDatabase = shelve.open(databasePath, "c")
        mutantTypeDatabase = dict()

        # go through each file, parse it, calculate all mutations, and generate files accordingly.
        for srcFile in javaRead.fileList:
            print("(" + str(fileCounter + 1) + "/" + str(fileCount) + ") source file: ", srcFile)
            targetList = list()

            try:
                # parsing the source file into a tree.
                tree = javaParse.parse(javaRead.getFileContent(srcFile))

                # assigning a number to each node to be able to identify it uniquely.
                javaParse.numerify(tree)
                # javaParse.tree2DOT(tree)

            except Exception as e:
                print("Error in parsing Java code, skipping the file.")
                sys.stderr.write(str(e))
                continue

            fileCounter += 1

            if options.isAll:
                enabledMutators = "all"
            elif options.isNullCheck:
                enabledMutators = "null-check"
            else:
                enabledMutators = "classical"

            # apply mutations on the tree and receive the resulting mutants as a list of strings, and a detailed
            # list of which operators created how many mutants.
            javaMutate.mutantsPerLine = dict()
            mutated, mutantTypes = javaMutate.applyMutators(tree, higherOrder, enabledMutators)

            print("--> mutations found: ", len(mutated))

            # go through all mutant types, and add them in total. also output the info to the user.
            for mutantType in list(mutantTypes.keys()):
                if mutantTypes[mutantType] > 0:
                    print("---->", mutantType, ":", mutantTypes[mutantType])
                mutantTypeDatabase[mutantType] = mutantTypes[mutantType] + mutantTypeDatabase.get(mutantType, 0)
            totalMutantCount += len(mutated)

            # for each mutant, generate the file, and add it to the list.
            for mutatedFile in mutated:
                targetList.append(javaRead.generateNewFile(srcFile, mutatedFile, javaMutate.mutantsPerLine))

            # if the list is not empty (some mutants were found), put the data in the database.
            if len(targetList) != 0:
                mutationDatabase[os.path.relpath(srcFile, javaRead.sourceDirectory)] = targetList

        mutationDatabase.close()

        print("total mutations found: ", totalMutantCount)
        for mutantType in list(mutantTypeDatabase.keys()):
            if mutantTypeDatabase[mutantType] > 0:
                print("-->", mutantType, ":", mutantTypeDatabase[mutantType])

    # *****************************************************************************************************************
    # ---------------------------------------- test suite running phase -----------------------------------------------
    # *****************************************************************************************************************

    if options.isBuildActive:

        # let's tell the user upfront that this may corrupt the source code.
        print("\n\n!!! CAUTION !!!")
        print("code can be changed accidentally. use a backup version.\n")

        reportGenerator = ReportGenerator(littleDarwinVersion)

        if options.alternateDb == "***dummy***":
            databasePath = os.path.abspath(os.path.join(options.buildPath, "LittleDarwinResults", "mutationdatabase"))
        else:
            databasePath = options.alternateDb

        mutantsPath = os.path.dirname(databasePath)
        assert os.path.isdir(mutantsPath)

        resultsDatabasePath = databasePath + "-results"
        reportGenerator.initiateDatabase(resultsDatabasePath)

        try:
            if os.path.basename(options.buildPath) == "pom.xml":
                assert os.path.isfile(options.buildPath)
                buildDir = os.path.abspath(os.path.dirname(options.buildPath))
            else:
                assert os.path.isdir(options.buildPath)
                buildDir = os.path.abspath(options.buildPath)

        except AssertionError as exception:
            print("build system working directory should be a directory.")

        # check if we have separate test-suite
        if options.testCommand != "***dummy***":
            separateTestSuite = True
            if options.testPath == "***dummy***":
                testDir = buildDir
            else:
                try:
                    if os.path.basename(options.buildPath) == "pom.xml":
                        assert os.path.isfile(options.buildPath)
                        testDir = os.path.abspath(os.path.dirname(options.testPath))
                    else:
                        assert os.path.isdir(options.buildPath)
                        testDir = os.path.abspath(options.testPath)

                except AssertionError as exception:
                    print("test project build system working directory should be a directory.")

        else:
            separateTestSuite = False

        # try to open the database. if it can't be opened, it means that it does not exist or it is corrupt.
        try:
            mutationDatabase = shelve.open(databasePath, "r")
        except:
            print("cannot open mutation database. it may be corrupted or unavailable. delete all generated files and run the mutant generation phase again.")
            sys.exit(1)

        databaseKeys = list(mutationDatabase.keys())
        assert isinstance(databaseKeys, list)

        # let's sort the mutants by name to create the possibility of following the flow of the process by user.
        databaseKeys.sort()

        # only here for debugging purposes
        # for desired in databaseKeys:
        #     if "PluginMap.java" in desired:
        #         desiredIndex = databaseKeys.index(desired)
        #         break
        #
        # databaseKeys.insert(0, databaseKeys.pop(desiredIndex))
        #

        mutationDatabaseLength = len(databaseKeys)
        textReportData = list()
        htmlReportData = list()
        fileCounter = 0

        # initial build check to avoid false results. the system must be able to build cleanly without errors.

        # use build command for the initial build unless it is explicitly provided.
        if options.initialBuildCommand == "***dummy***":
            commandString = options.buildCommand.split(',')
        else:
            commandString = options.initialBuildCommand.split(',')

        print("Initial build... ",)

        try:
            processKilled, processExitCode, initialOutput = timeoutAlternative(commandString,
                                                                               workingDirectory=buildDir,
                                                                               timeout=int(options.timeout))

            # initialOutput = subprocess.check_output(commandString, stderr=subprocess.STDOUT, cwd=buildDir)
            # workaround for older python versions
            if processKilled or processExitCode:
                raise subprocess.CalledProcessError(1 if processKilled else processExitCode, commandString,
                                                    initialOutput)

            with open(os.path.abspath(os.path.join(mutantsPath, "initialbuild.txt")), 'w') as contentFile:
                contentFile.write(str(initialOutput))
            print("done.\n\n")

        except subprocess.CalledProcessError as exception:
            initialOutput = exception.output
            with open(os.path.abspath(os.path.join(mutantsPath, "initialbuild.txt")), 'w') as contentFile:
                contentFile.write(str(initialOutput))

            print("failed.\n")
            print("Initial build failed. Try building the system manually first to make sure it can be built. Take a look at " + os.path.abspath(
                os.path.join(mutantsPath, "initialbuild.txt")) + " to find out why this happened.")
            sys.exit(1)

        totalMutantCount = 0
        totalMutantCounter = 0

        for key in databaseKeys:
            totalMutantCount += len(mutationDatabase[key])

        startTime = time.time()

        # running the build system for each mutant.
        for key in databaseKeys:

            fileCounter += 1

            print("(" + str(fileCounter) + "/" + str(mutationDatabaseLength) + ") collecting results for ", key)

            mutantCount = len(mutationDatabase[key])
            mutantCounter = 0

            successList = list()
            failureList = list()

            # for each mutant, replace the original file, run the build, store the results
            for replacementFileRel in mutationDatabase[key]:
                replacementFile = os.path.abspath(os.path.join(mutantsPath, replacementFileRel))
                mutantCounter += 1
                totalMutantCounter += 1

                # let's make sure that runOutput is empty, and not None to begin with.
                runOutput = bytes()
                runOutputTest = bytes()

                # replace the original file with the mutant
                shutil.copyfile(replacementFile, os.path.join(options.sourcePath, key))

                commandString = options.buildCommand.split(',')
                if separateTestSuite:
                    testCommandString = options.testCommand.split(',')

                try:
                    # if we have timeout support, simply run the command with timeout support from subprocess32
                    # if timeoutSupport:
                    #     runOutput = subprocess.check_output(commandString, stderr=subprocess.STDOUT, cwd=buildDir,
                    #                                         timeout=int(options.timeout))
                    #     if separateTestSuite:
                    #         runOutput += subprocess.check_output(testCommandString, stderr=subprocess.STDOUT, cwd=testDir,
                    #                                         timeout=int(options.timeout))

                    # else, run our alternative method
                    # else:
                    processKilled, processExitCode, runOutput = timeoutAlternative(commandString,
                                                                                   workingDirectory=buildDir,
                                                                                   timeout=int(options.timeout))

                    # raise the same exception as the original check_output.
                    if processKilled or processExitCode:
                        raise subprocess.CalledProcessError(1 if processKilled else processExitCode, commandString,
                                                            runOutput)

                    if separateTestSuite:
                        processKilled, processExitCode, runOutputTest = timeoutAlternative(testCommandString,
                                                                                           workingDirectory=testDir,
                                                                                           timeout=int(options.timeout))

                        # raise the same exception as the original check_output.
                        if processKilled or processExitCode:
                            raise subprocess.CalledProcessError(1 if processKilled else processExitCode,
                                                                commandString, "\n".join([runOutput, runOutputTest]))

                    # if we are here, it means no exceptions happened, so lets add this to our success list.
                    runOutput = runOutput.decode("utf-8") + '\n' + runOutputTest.decode("utf-8")
                    successList.append(os.path.basename(replacementFile))

                # putting two exceptions in one except clause, specially when one of them is not defined on some
                # platforms does not look like a good idea; even though both of them do exactly the same thing.
                except subprocess.CalledProcessError as exception:
                    runOutput = exception.output.decode("utf-8")
                    # oops, error. let's add this to failure list.
                    failureList.append(os.path.basename(replacementFile))

                # except subprocess.TimeoutExpired as exception:
                #     runOutput = exception.output
                #     failureList.append(os.path.basename(replacementFile))

                targetTextOutputFile = os.path.splitext(replacementFile)[0] + ".txt"

                # we can't use print, since we like to write on the same line again.
                sys.stdout.write(
                    "elapsed: " + str(datetime.timedelta(seconds=int(time.time() - startTime))) + " remaining: " + str(
                        datetime.timedelta(seconds=int((old_div(float(time.time() - startTime), totalMutantCounter)) * float(
                            totalMutantCount - totalMutantCounter)))) + " total: " + str(
                        totalMutantCounter) + "/" + str(totalMutantCount) + " current: " + str(
                        mutantCounter) + "/" + str(mutantCount) + " *** survived: " + str(
                        len(successList)) + " - killed: " + str(len(failureList)) + "         \r")
                sys.stdout.flush()

                # writing the build output to disk.
                with open(targetTextOutputFile, 'w') as contentFile:
                    contentFile.write(str(runOutput))

                # if there's a cleanup option, execute it. the results will be ignored because we don't want our process
                #  to be interrupted if there's nothing to clean up.
                if options.cleanUp != "***dummy***":
                    subprocess.call(options.cleanUp.split(","), cwd=buildDir)
                    if separateTestSuite:
                        subprocess.call(options.cleanUp.split(","), cwd=testDir)

                # workaround:
                # shutil.rmtree(os.path.join(testDir,"VolumetryLoggerTest"),ignore_errors=True)

            # all mutants must be checked by now, so we should have a complete divide between success and failure.
            assert len(successList) + len(failureList) == mutantCount

            # append the information for this file to the reports.
            textReportData.append(key + ": survived (" + str(len(successList)) + "/" + str(mutantCount) + ") -> " + str(
                successList) + " - killed (" + str(len(failureList)) + "/" + str(mutantCount) + ") -> " + str(
                failureList) + "\r\n")
            htmlReportData.append([key, len(successList), mutantCount])

            # we are done with the file. let's return it to the original state.
            shutil.copyfile(os.path.join(os.path.dirname(replacementFile), "original.java"),
                            os.path.join(options.sourcePath, key))

            # generate an HTML report for the file.

            targetHTMLOutputFile = os.path.join(os.path.dirname(replacementFile), "results.html")
            with open(targetHTMLOutputFile, 'w') as contentFile:
                contentFile.write(
                    reportGenerator.generateHTMLReportPerFile(key, targetHTMLOutputFile, successList, failureList))

            print("\n\n")

        # write final text report.
        with open(os.path.abspath(os.path.join(mutantsPath, "report.txt")),
                  'w') as textReportFile:
            textReportFile.writelines(textReportData)

        # write final HTML report.
        targetHTMLReportFile = os.path.abspath(
            os.path.join(mutantsPath, "report.html"))
        with open(targetHTMLReportFile, 'w') as htmlReportFile:
            htmlReportFile.writelines(reportGenerator.generateHTMLFinalReport(htmlReportData, targetHTMLReportFile))

    # if neither build nor mutation phase is active, let's help the user.
    if not (options.isBuildActive or options.isMutationActive):
        optionParser.print_help()
        print("\nExample:\n  LittleDarwin -m -b -t ./ -p ./src/main -c mvn,clean,test --timeout=120\n\n")





#
# [The "BSD license"]
#  Copyright (c) 2012 Terence Parr
#  Copyright (c) 2012 Sam Harwell
#  Copyright (c) 2014 Eric Vergnaud
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#  3. The name of the author may not be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
#  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
#  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
#  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
#  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
#  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
#  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


#
# This implementation of {@link ANTLRErrorListener} can be used to identify
# certain potential correctness and performance problems in grammars. "Reports"
# are made by calling {@link Parser#notifyErrorListeners} with the appropriate
# message.
#
# <ul>
# <li><b>Ambiguities</b>: These are cases where more than one path through the
# grammar can match the input.</li>
# <li><b>Weak context sensitivity</b>: These are cases where full-context
# prediction resolved an SLL conflict to a unique alternative which equaled the
# minimum alternative of the SLL conflict.</li>
# <li><b>Strong (forced) context sensitivity</b>: These are cases where the
# full-context prediction resolved an SLL conflict to a unique alternative,
# <em>and</em> the minimum alternative of the SLL conflict was found to not be
# a truly viable alternative. Two-stage parsing cannot be used for inputs where
# this situation occurs.</li>
# </ul>

from builtins import str
from io import StringIO
from custom_antlr4.Utils import str_set
from custom_antlr4.error.ErrorListener import ErrorListener

class DiagnosticErrorListener(ErrorListener):

    def __init__(self, exactOnly=True):
        # whether all ambiguities or only exact ambiguities are reported.
        self.exactOnly = exactOnly

    def reportAmbiguity(self, recognizer, dfa, startIndex,
                       stopIndex, exact, ambigAlts, configs):
        if self.exactOnly and not exact:
            return

        with StringIO() as buf:
            buf.write(u"reportAmbiguity d=")
            buf.write(self.getDecisionDescription(recognizer, dfa))
            buf.write(u": ambigAlts=")
            buf.write(str_set(self.getConflictingAlts(ambigAlts, configs)))
            buf.write(u", input='")
            buf.write(recognizer.getTokenStream().getText((startIndex, stopIndex)))
            buf.write(u"'")
            recognizer.notifyErrorListeners(buf.getvalue())


    def reportAttemptingFullContext(self, recognizer, dfa, startIndex,
                       stopIndex, conflictingAlts, configs):
        with StringIO() as buf:
            buf.write(u"reportAttemptingFullContext d=")
            buf.write(self.getDecisionDescription(recognizer, dfa))
            buf.write(u", input='")
            buf.write(recognizer.getTokenStream().getText((startIndex, stopIndex)))
            buf.write(u"'")
            recognizer.notifyErrorListeners(buf.getvalue())

    def reportContextSensitivity(self, recognizer, dfa, startIndex,
                       stopIndex, prediction, configs):
        with StringIO() as buf:
            buf.write(u"reportContextSensitivity d=")
            buf.write(self.getDecisionDescription(recognizer, dfa))
            buf.write(u", input='")
            buf.write(recognizer.getTokenStream().getText((startIndex, stopIndex)))
            buf.write(u"'")
            recognizer.notifyErrorListeners(buf.getvalue())

    def getDecisionDescription(self, recognizer, dfa):
        decision = dfa.decision
        ruleIndex = dfa.atnStartState.ruleIndex

        ruleNames = recognizer.ruleNames
        if ruleIndex < 0 or ruleIndex >= len(ruleNames):
            return str(decision)

        ruleName = ruleNames[ruleIndex]
        if ruleName is None or len(ruleName)==0:
            return str(decision)

        return str(decision) + u" (" + ruleName + u")"

    #
    # Computes the set of conflicting or ambiguous alternatives from a
    # configuration set, if that information was not already provided by the
    # parser.
    #
    # @param reportedAlts The set of conflicting or ambiguous alternatives, as
    # reported by the parser.
    # @param configs The conflicting or ambiguous configuration set.
    # @return Returns {@code reportedAlts} if it is not {@code null}, otherwise
    # returns the set of alternatives represented in {@code configs}.
    #
    def getConflictingAlts(self, reportedAlts, configs):
        if reportedAlts is not None:
            return reportedAlts

        result = set()
        for config in configs:
            result.add(config.alt)

        return result



#
# [The "BSD license"]
#  Copyright (c) 2012 Terence Parr
#  Copyright (c) 2012 Sam Harwell
#  Copyright (c) 2014 Eric Vergnaud
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#  3. The name of the author may not be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
#  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
#  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
#  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
#  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
#  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
#  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


# A set of utility routines useful for all kinds of ANTLR trees.#
from builtins import str
from builtins import range
from builtins import object
from io import StringIO
from custom_antlr4.Token import Token
from custom_antlr4.Utils import escapeWhitespace
from custom_antlr4.tree.Tree import RuleNode, ErrorNode, TerminalNode

class Trees(object):

     # Print out a whole tree in LISP form. {@link #getNodeText} is used on the
    #  node payloads to get the text for the nodes.  Detect
    #  parse trees and extract data appropriately.
    @classmethod
    def toStringTree(cls, t, ruleNames=None, recog=None):
        if recog is not None:
            ruleNames = recog.ruleNames
        s = escapeWhitespace(cls.getNodeText(t, ruleNames), False)
        if t.getChildCount()==0:
            return s
        with StringIO() as buf:
            buf.write(u"(")
            buf.write(s)
            buf.write(u' ')
            for i in range(0, t.getChildCount()):
                if i > 0:
                    buf.write(u' ')
                buf.write(cls.toStringTree(t.getChild(i), ruleNames))
            buf.write(u")")
            return buf.getvalue()

    @classmethod
    def getNodeText(cls, t, ruleNames=None, recog=None):
        if recog is not None:
            ruleNames = recog.ruleNames
        if ruleNames is not None:
            if isinstance(t, RuleNode):
                return ruleNames[t.getRuleContext().getRuleIndex()]
            elif isinstance( t, ErrorNode):
                return str(t)
            elif isinstance(t, TerminalNode):
                if t.symbol is not None:
                    return t.symbol.text
        # no recog for rule names
        payload = t.getPayload()
        if isinstance(payload, Token ):
            return payload.text
        return str(t.getPayload())


    # Return ordered list of all children of this node
    @classmethod
    def getChildren(cls, t):
        return [ t.getChild(i) for i in range(0, t.getChildCount()) ]

    # Return a list of all ancestors of this node.  The first node of
    #  list is the root and the last is the parent of this node.
    #
    @classmethod
    def getAncestors(cls, t):
        ancestors = []
        t = t.getParent()
        while t is not None:
            ancestors.append(0, t) # insert at start
            t = t.getParent()
        return ancestors

    @classmethod
    def findAllTokenNodes(cls, t, ttype):
        return cls.findAllNodes(t, ttype, True)

    @classmethod
    def findAllRuleNodes(cls, t, ruleIndex):
        return cls.findAllNodes(t, ruleIndex, False)

    @classmethod
    def findAllNodes(cls, t, index, findTokens):
        nodes = []
        cls._findAllNodes(t, index, findTokens, nodes)
        return nodes

    @classmethod
    def _findAllNodes(cls, t, index, findTokens, nodes):
        from custom_antlr4.ParserRuleContext import ParserRuleContext
        # check this node (the root) first
        if findTokens and isinstance(t, TerminalNode):
            if t.symbol.type==index:
                nodes.append(t)
        elif not findTokens and isinstance(t, ParserRuleContext):
            if t.ruleIndex == index:
                nodes.append(t)
        # check children
        for i in range(0, t.getChildCount()):
            cls._findAllNodes(t.getChild(i), index, findTokens, nodes)

    @classmethod
    def descendants(cls, t):
        nodes = []
        nodes.append(t)
        for i in range(0, t.getChildCount()):
            nodes.extend(cls.descendants(t.getChild(i)))
        return nodes



from __future__ import print_function
from builtins import str
from builtins import range
from builtins import object
import sys
from littledarwin.JavaParser import JavaParser
from littledarwin.JavaParse import JavaParse
import copy

from custom_antlr4 import Token
from custom_antlr4.tree.Tree import TerminalNodeImpl
from random import shuffle
from math import log10

sys.setrecursionlimit(50000)


# TODO: Every mutation operator must be implemented using this class in the future
# ---------------------------------------------------------------------------------------
class MutationOperator(object):
    def __init__(self):
        self.nodeIndexes = list()
        self.originalNodes = dict()
        self.mutatedNodes = dict()
        self.type = None
        self.tree = None

    def seekNodes(self):
        pass

    def filterNodes(self):
        pass

    def mutateNode(self, nodeIndex):
        pass

    def unmutateNode(self, nodeIndex):
        pass

    def getTree(self):
        assert isinstance(self.tree, JavaParser.CompilationUnitContext)
        return self.tree

    def getTexts(self):
        assert isinstance(self.tree, JavaParser.CompilationUnitContext)

        mutatedCode = list()

        for nodeIndex in self.nodeIndexes:
            mutationBefore = "----> before: " + self.originalNodes[nodeIndex].getText()
            mutationAfter = "----> after: " + self.mutatedNodes[nodeIndex].getText()

            self.mutateNode(nodeIndex)
            mutatedCode.append((
                    "/* LittleDarwin generated mutant \nmutant type: " + self.type + "\n" + mutationBefore + "\n" + mutationAfter + "\n----> line number in original file: " + str(
                self.mutatedNodes[nodeIndex].start.line) + "\n----> mutated nodes: " + str(nodeIndex) + "\n*/ \n\n" + (
                        " ".join(self.tree.getText().rsplit("<EOF>", 1)))))
            self.unmutateNode(nodeIndex)

        return mutatedCode

    def getNodes(self):
        return self.nodeIndexes

    def setTree(self, tree):
        assert isinstance(tree, JavaParser.CompilationUnitContext)
        self.tree = tree


# ---------------------------------------------------------------------------------------


class JavaMutate(object):
    def __init__(self, javaParseObjectInput=None, verbose=False):
        self.verbose = verbose
        self.mutantsPerLine = dict()

        if javaParseObjectInput is not None:
            self.javaParseObject = javaParseObjectInput
        else:
            self.javaParseObject = JavaParse(self.verbose)

    def returnNodeToOriginalState(self, node):
        assert isinstance(node, TerminalNodeImpl)

        try:
            if node.symbol.oldText is not None:
                node.symbol.text = node.symbol.oldText
        except AttributeError as e:
            pass

    def returnTreeToOriginalState(self, tree):
        assert isinstance(tree, JavaParser.CompilationUnitContext)

        stack = list()
        stack.append(tree)

        while len(stack) > 0:
            currentNode = stack.pop()

            if isinstance(currentNode, TerminalNodeImpl):
                self.returnNodeToOriginalState(currentNode)

            if currentNode.getChildCount() > 0:
                stack.extend(currentNode.children)

    def findMutableNodes(self, tree):
        assert isinstance(tree, JavaParser.CompilationUnitContext)

        mutableNodes = list()
        mutableNodes.extend(self.arithmeticOperatorReplacementBinary(tree, "return_nodes"))
        mutableNodes.extend(self.arithmeticOperatorReplacementShortcut(tree, "return_nodes"))
        mutableNodes.extend(self.arithmeticOperatorReplacementUnary(tree, "return_nodes"))
        mutableNodes.extend(self.assignmentOperatorReplacementShortcut(tree, "return_nodes"))
        mutableNodes.extend(self.conditionalOperatorDeletion(tree, "return_nodes"))
        mutableNodes.extend(self.conditionalOperatorReplacement(tree, "return_nodes"))
        mutableNodes.extend(self.logicalOperatorReplacement(tree, "return_nodes"))
        mutableNodes.extend(self.relationalOperatorReplacement(tree, "return_nodes"))
        mutableNodes.extend(self.shiftOperatorReplacement(tree, "return_nodes"))

        return mutableNodes

    def runHigherOrderProcedure(self, tree, nodeGroup):
        assert isinstance(tree, JavaParser.CompilationUnitContext)
        assert isinstance(nodeGroup, list)
        steps = len(nodeGroup)
        assert steps >= 1

        # Store original nodes here
        # originalNodes = list()

        originalNodes = [node[0] for node in nodeGroup]

        # for node in nodeGroup:
        # originalNodeAddress = self.javaParseObject.getNode(tree, node[0])
        # originalNodeContent = copy.deepcopy(originalNodeAddress)
        # originalNodes.append(node)

        while steps >= 1:
            steps -= 1

            node = nodeGroup.pop()

            if node[1] == "ArithmeticOperatorReplacementBinary":
                tree = self.arithmeticOperatorReplacementBinary(tree, "return_tree", node[0])
            elif node[1] == "ArithmeticOperatorReplacementShortcut":
                tree = self.arithmeticOperatorReplacementShortcut(tree, "return_tree", node[0])
            elif node[1] == "ArithmeticOperatorReplacementUnary":
                tree = self.arithmeticOperatorReplacementUnary(tree, "return_tree", node[0])
            elif node[1] == "AssignmentOperatorReplacementShortcut":
                tree = self.assignmentOperatorReplacementShortcut(tree, "return_tree", node[0])
            elif node[1] == "ConditionalOperatorDeletion":
                tree = self.conditionalOperatorDeletion(tree, "return_tree", node[0])
            elif node[1] == "ConditionalOperatorReplacement":
                tree = self.conditionalOperatorReplacement(tree, "return_tree", node[0])
            elif node[1] == "LogicalOperatorReplacement":
                tree = self.logicalOperatorReplacement(tree, "return_tree", node[0])
            elif node[1] == "RelationalOperatorReplacement":
                tree = self.relationalOperatorReplacement(tree, "return_tree", node[0])
            elif node[1] == "ShiftOperatorReplacement":
                tree = self.shiftOperatorReplacement(tree, "return_tree", node[0])

        assert len(nodeGroup) == 0

        mutatedText = "/* LittleDarwin generated higher order mutant\n ----> line number in original file: " + str(
            [self.javaParseObject.getNode(tree, node).start.line for node in
             originalNodes]).strip("[]") + "\n----> mutated nodes: " + str(originalNodes).strip("[]") + "\n*/ \n\n" + (
                          " ".join(tree.getText().rsplit("<EOF>", 1)))  # create compilable, readable code

        # for node in originalNodes:
        #     self.javaParseObject.setNode(tree, node[0], node[1])

        self.returnTreeToOriginalState(tree)

        return mutatedText

    def applyHigherOrderMutatorsBasedOnDistance(self, tree, higherOrderDirective):
        assert isinstance(tree, JavaParser.CompilationUnitContext)
        print("Finding mutable nodes in AST")
        # Find all mutable nodes
        nodes = self.findMutableNodes(tree)
        if len(nodes) == 0:
            return list()

        # Combine all nodes RANDOMLY
        print("Shuffling Mutants")
        shuffle(nodes)
        nodeGroups = list()

        higherOrder = int(2 * log10(len(nodes))) if higherOrderDirective == -1 else higherOrderDirective
        higherOrder = 1 if higherOrder == 0 else higherOrder

        print("Preparing Higher-Order Mutants")
        for i in range(0, len(nodes) - higherOrder + 1, higherOrder):
            nodeGroups.append([nodes[j] for j in range(i, i + higherOrder, 1)])

        # Run mutation operators

        mutatedTreeTexts = list()
        c = 0

        for nodeGroup in nodeGroups:
            c += 1
            sys.stdout.write("\rProcessing HOM " + str(c) + "/" + str(len(nodeGroups)))
            sys.stdout.flush()

            mutatedTreeTexts.append(self.runHigherOrderProcedure(tree, nodeGroup))

        print(" ")

        return mutatedTreeTexts

    def applyHigherOrderMutators(self, tree, higherOrderDirective):
        assert isinstance(tree, JavaParser.CompilationUnitContext)
        print("Finding mutable nodes in AST")
        # Find all mutable nodes
        nodes = self.findMutableNodes(tree)
        if len(nodes) == 0:
            return list()

        # Combine all nodes RANDOMLY
        print("Shuffling Mutants")
        shuffle(nodes)
        nodeGroups = list()

        higherOrder = int(2 * log10(len(nodes))) if higherOrderDirective == -1 else higherOrderDirective
        higherOrder = 1 if higherOrder == 0 else higherOrder

        print("Preparing Higher-Order Mutants")
        for i in range(0, len(nodes) - higherOrder + 1, higherOrder):
            nodeGroups.append([nodes[j] for j in range(i, i + higherOrder, 1)])

        # Run mutation operators

        mutatedTreeTexts = list()
        c = 0

        for nodeGroup in nodeGroups:
            c += 1
            sys.stdout.write("\rProcessing HOM " + str(c) + "/" + str(len(nodeGroups)))
            sys.stdout.flush()

            mutatedTreeTexts.append(self.runHigherOrderProcedure(tree, nodeGroup))

        print(" ")

        return mutatedTreeTexts

    def applyMutators(self, tree, higherOrder, type):
        assert isinstance(tree, JavaParser.CompilationUnitContext)

        mutationTypeCount = dict()

        mutationTypeCount["ArithmeticOperatorReplacementBinary"] = 0
        mutationTypeCount["ArithmeticOperatorReplacementShortcut"] = 0
        mutationTypeCount["ArithmeticOperatorReplacementUnary"] = 0
        mutationTypeCount["LogicalOperatorReplacement"] = 0
        mutationTypeCount["ShiftOperatorReplacement"] = 0
        mutationTypeCount["RelationalOperatorReplacement"] = 0
        mutationTypeCount["ConditionalOperatorReplacement"] = 0
        mutationTypeCount["ConditionalOperatorDeletion"] = 0
        mutationTypeCount["AssignmentOperatorReplacementShortcut"] = 0
        mutationTypeCount["NullifyInputVariable"] = 0
        mutationTypeCount["RemoveNullCheck"] = 0
        mutationTypeCount["NullifyReturnValue"] = 0
        mutationTypeCount["Higher-Order"] = 0
        mutationTypeCount["NullifyObjectInitialization"] = 0

        if higherOrder > 1 or higherOrder == -1:
            treeTexts = self.applyHigherOrderMutators(tree, higherOrder)
            mutationTypeCount["Higher-Order"] = len(treeTexts)

            return treeTexts, mutationTypeCount

        mutatedTrees = list()

        # mutatedTrees.extend(self.arithmeticOperatorDeletionShortcut(tree))
        # mutatedTrees.extend(self.arithmeticOperatorDeletionUnary(tree))
        # mutatedTrees.extend(self.arithmeticOperatorInsertionShortcut(tree))
        # mutatedTrees.extend(self.arithmeticOperatorInsertionUnary(tree))
        # mutatedTrees.extend(self.conditionalOperatorInsertion(tree))
        # mutatedTrees.extend(self.logicalOperatorDeletion(tree))
        # mutatedTrees.extend(self.logicalOperatorInsertion(tree))

        if type == "classical" or type == "all":
            resultArithmeticOperatorReplacementBinary = self.arithmeticOperatorReplacementBinary(tree)
            mutatedTrees.extend(resultArithmeticOperatorReplacementBinary)
            mutationTypeCount["ArithmeticOperatorReplacementBinary"] = len(resultArithmeticOperatorReplacementBinary)

            resultArithmeticOperatorReplacementShortcut = self.arithmeticOperatorReplacementShortcut(tree)
            mutatedTrees.extend(resultArithmeticOperatorReplacementShortcut)
            mutationTypeCount["ArithmeticOperatorReplacementShortcut"] = len(
                resultArithmeticOperatorReplacementShortcut)

            resultArithmeticOperatorReplacementUnary = self.arithmeticOperatorReplacementUnary(tree)
            mutatedTrees.extend(resultArithmeticOperatorReplacementUnary)
            mutationTypeCount["ArithmeticOperatorReplacementUnary"] = len(resultArithmeticOperatorReplacementUnary)

            # currently generating too many mutations
            # resultNegateConditionalsMutator = self.negateConditionalsMutator(tree)
            # mutatedTrees.extend(resultNegateConditionalsMutator)
            # mutationTypeCount["NegateConditionalsMutator"] = len(resultNegateConditionalsMutator)

            resultLogicalOperatorReplacement = self.logicalOperatorReplacement(tree)
            mutatedTrees.extend(resultLogicalOperatorReplacement)
            mutationTypeCount["LogicalOperatorReplacement"] = len(resultLogicalOperatorReplacement)

            resultShiftOperatorReplacement = self.shiftOperatorReplacement(tree)
            mutatedTrees.extend(resultShiftOperatorReplacement)
            mutationTypeCount["ShiftOperatorReplacement"] = len(resultShiftOperatorReplacement)

            resultRelationalOperatorReplacement = self.relationalOperatorReplacement(tree)
            mutatedTrees.extend(resultRelationalOperatorReplacement)
            mutationTypeCount["RelationalOperatorReplacement"] = len(resultRelationalOperatorReplacement)

            resultConditionalOperatorReplacement = self.conditionalOperatorReplacement(tree)
            mutatedTrees.extend(resultConditionalOperatorReplacement)
            mutationTypeCount["ConditionalOperatorReplacement"] = len(resultConditionalOperatorReplacement)

            resultConditionalOperatorDeletion = self.conditionalOperatorDeletion(tree)
            mutatedTrees.extend(resultConditionalOperatorDeletion)
            mutationTypeCount["ConditionalOperatorDeletion"] = len(resultConditionalOperatorDeletion)

            resultAssignmentOperatorReplacementShortcut = self.assignmentOperatorReplacementShortcut(tree)
            mutatedTrees.extend(resultAssignmentOperatorReplacementShortcut)
            mutationTypeCount["AssignmentOperatorReplacementShortcut"] = len(
                resultAssignmentOperatorReplacementShortcut)

        if type == "object-oriented" or type == "all":
            pass

        if type == "null-check" or type == "all":
            resultNullifyInputVariable = self.nullifyInputVariable(tree)
            mutatedTrees.extend(resultNullifyInputVariable)
            mutationTypeCount["NullifyInputVariable"] = len(resultNullifyInputVariable)

            resultNullifyReturnValue = self.nullifyReturnValue(tree)
            mutatedTrees.extend(resultNullifyReturnValue)
            mutationTypeCount["NullifyReturnValue"] = len(resultNullifyReturnValue)

            resultNullifyObjectInitialization = self.nullifyObjectInitialization(tree)
            mutatedTrees.extend(resultNullifyObjectInitialization)
            mutationTypeCount["NullifyObjectInitialization"] = len(resultNullifyObjectInitialization)

            # if type == "null-check":
            resultRemoveNullCheck = self.removeNullCheck(tree)
            mutatedTrees.extend(resultRemoveNullCheck)
            mutationTypeCount["RemoveNullCheck"] = len(resultRemoveNullCheck)

        return mutatedTrees, mutationTypeCount

    """
    Mutation operator guidelines:
    input: tree, mode, single node index
    output: tree, text, node list

    ----- mode: return_text
                return_nodes
                return_tree
    """

    def removeNullCheck(self, tree, mode="return_text", nodeIndex=None):
        assert isinstance(tree, JavaParser.CompilationUnitContext)

        if mode == "return_text":
            mutatedTreesTexts = list()
            expressionList = self.javaParseObject.seek(tree, JavaParser.ExpressionContext)

            while len(expressionList) > 0:
                # tmpTree = copy.deepcopy(tree)
                expressionIndex = expressionList.pop()

                node = self.javaParseObject.getNode(tree, expressionIndex)
                assert isinstance(node, JavaParser.ExpressionContext)

                try:
                    if not (isinstance(node.children[0], JavaParser.ExpressionContext) and
                            isinstance(node.children[1], TerminalNodeImpl) and
                            isinstance(node.children[2], JavaParser.ExpressionContext)):
                        continue  # not a binary expression
                except Exception as e:
                    continue

                if not (node.children[1].symbol.text == u"!=" or node.children[1].symbol.text == u"=="):
                    continue  # not a relational operator

                if not u'null' in node.getText():
                    continue  # not a null check

                mutationBefore = "----> before: " + node.getText()
                if self.verbose:
                    print(mutationBefore)

                originalText = copy.deepcopy(node.children[1].symbol.text)

                if originalText == u"==":
                    node.children[1].symbol.text = u"!="
                elif originalText == u"!=":
                    node.children[1].symbol.text = u"=="
                else:
                    assert False

                mutationAfter = "----> after: " + node.getText()

                if self.verbose:
                    print(mutationAfter)

                mutatedTreesTexts.append((
                    (
                            "/* LittleDarwin generated mutant\nmutant type: removeNullCheck\n" + mutationBefore + "\n" + mutationAfter + "\n----> line number in original file: " + str(
                        node.start.line) + "\n----> mutated nodes: " + str(expressionIndex) + "\n*/ \n\n" + (
                                " ".join(tree.getText().rsplit("<EOF>", 1))))))  # create compilable, readable code

                node.children[1].symbol.text = copy.deepcopy(originalText)

                if node.start.line in self.mutantsPerLine.keys():
                    self.mutantsPerLine[node.start.line] += 1
                else:
                    self.mutantsPerLine[node.start.line] = 1

            return mutatedTreesTexts

    # def removeNullCheck(self, tree, mode="return_text", nodeIndex=None):
    #     assert isinstance(tree, JavaParser.CompilationUnitContext)
    #
    #     if mode == "return_text":
    #
    #         mutatedTreesTexts = list()
    #         expressionList = self.javaParseObject.seek(tree, JavaParser.ParExpressionContext)
    #
    #         while len(expressionList) > 0:
    #             # tmpTree = copy.deepcopy(tree)
    #             expressionIndex = expressionList.pop()
    #
    #             node = self.javaParseObject.getNode(tree, expressionIndex)
    #             assert isinstance(node, JavaParser.ParExpressionContext)
    #             assert isinstance(node.children[0], TerminalNodeImpl)
    #             assert isinstance(node.children[2], TerminalNodeImpl)
    #
    #             if u'null' not in node.getText():
    #                 continue
    #
    #             if u'==' in node.getText():
    #                 replacementText = u'false'
    #             elif u'!=' in node.getText():
    #                 replacementText = u'true'
    #             else:
    #                 continue
    #
    #             mutationBefore = "----> before: " + node.getText()
    #             if self.verbose:
    #                 print mutationBefore
    #             originalText0 = copy.deepcopy(node.children[0].symbol.text)
    #             node.children[0].symbol.text = u"(" + replacementText + u" /* "
    #             originalText2 = copy.deepcopy(node.children[2].symbol.text)
    #             node.children[2].symbol.text = u" */ )"
    #             mutationAfter = "----> after: " + node.getText().split(u'/*')[0]
    #             if self.verbose:
    #                 print mutationAfter
    #             mutatedTreesTexts.append((
    #                 "/* LittleDarwin generated mutant\nmutant type: removeNullCheck\n" + mutationBefore + "\n" + mutationAfter + "\n----> line number in original file: " + str(
    #                     node.start.line) + "\n*/ \n\n" + (
    #                     " ".join(tree.getText().rsplit("<EOF>", 1)))))
    #             # mutatedTreesTexts.append(tree.getText())
    #             node.children[0].symbol.text = copy.deepcopy(originalText0)
    #             node.children[2].symbol.text = copy.deepcopy(originalText2)
    #
    #         return mutatedTreesTexts

    # elif mode == "return_nodes":
    #     nodeList = list()
    #     expressionList = self.javaParseObject.seek(tree, JavaParser.ParExpressionContext)
    #
    #     for exp in expressionList:
    #         nodeList.append([exp, "negateConditionalsMutator"])

    def nullifyObjectInitialization(self, tree, mode="return_text", nodeIndex=None):
        assert isinstance(tree, JavaParser.CompilationUnitContext)

        if mode == "return_text":

            mutatedTreesTexts = list()
            expressionList = self.javaParseObject.seek(tree, JavaParser.CreatorContext)

            while len(expressionList) > 0:
                # tmpTree = copy.deepcopy(tree)
                expressionIndex = expressionList.pop()

                node = self.javaParseObject.getNode(tree, expressionIndex)
                assert isinstance(node, JavaParser.CreatorContext)
                try:
                    newStatement = node.parentCtx.getChild(0, TerminalNodeImpl)
                    argumentsStatement = node.children[-1].children[-1]

                    if newStatement.symbol.text != u'new':
                        continue

                    if not isinstance(argumentsStatement, JavaParser.ArgumentsContext):
                        continue

                    if argumentsStatement.children[-1].symbol.text != u')':
                        continue

                except:
                    continue

                mutationBefore = "----> before: " + node.parentCtx.getText()
                if self.verbose:
                    print(mutationBefore)
                originalText0 = copy.deepcopy(newStatement.symbol.text)
                newStatement.symbol.text = u" null /* "
                originalText2 = copy.deepcopy(argumentsStatement.children[-1].symbol.text)
                argumentsStatement.children[-1].symbol.text = u" ) */ "
                mutationAfter = "----> after: " + node.parentCtx.getText().split(u'/*')[0]

                if self.verbose:
                    print(mutationAfter)
                mutatedTreesTexts.append((
                        "/* LittleDarwin generated mutant\nmutant type: nullifyObjectInitialization\n" + mutationBefore + "\n" + mutationAfter + "\n----> line number in original file: " + str(
                    node.parentCtx.start.line) + "\n*/ \n\n" + (
                            " ".join(tree.getText().rsplit("<EOF>", 1)))))
                # mutatedTreesTexts.append(tree.getText())
                newStatement.symbol.text = copy.deepcopy(originalText0)
                argumentsStatement.children[-1].symbol.text = copy.deepcopy(originalText2)

                if node.start.line in self.mutantsPerLine.keys():
                    self.mutantsPerLine[node.start.line] += 1
                else:
                    self.mutantsPerLine[node.start.line] = 1

            return mutatedTreesTexts

    def nullifyReturnValue(self, tree, mode="return_text", nodeIndex=None):
        assert isinstance(tree, JavaParser.CompilationUnitContext)

        if mode == "return_text":

            mutatedTreesTexts = list()
            expressionList = self.javaParseObject.seek(tree, TerminalNodeImpl)

            while len(expressionList) > 0:
                # tmpTree = copy.deepcopy(tree)
                expressionIndex = expressionList.pop()

                node = self.javaParseObject.getNode(tree, expressionIndex)
                assert isinstance(node, TerminalNodeImpl)

                if u'return' != node.symbol.text:
                    continue

                if not isinstance(node.getParent().getChild(1), JavaParser.ExpressionContext):
                    continue

                parentMethod = self.javaParseObject.seekFirstMatchingParent(node, JavaParser.MethodDeclarationContext)
                if parentMethod is None:
                    continue

                assert isinstance(parentMethod, JavaParser.MethodDeclarationContext)

                parentType = parentMethod.getChild(0, JavaParser.JTypeContext)
                if not isinstance(parentType, JavaParser.JTypeContext):
                    continue

                if parentType.getChild(0, JavaParser.PrimitiveTypeContext) is not None:
                    continue  # primitive typed method

                mutationBefore = "----> before: " + node.getParent().getText()
                assert isinstance(node.symbol, Token)
                if self.verbose:
                    print(mutationBefore)
                originalText0 = copy.deepcopy(node.symbol.text)
                node.symbol.text = u"return null /* "
                originalText2 = copy.deepcopy(node.getParent().getChild(2).symbol.text)
                node.getParent().getChild(2).symbol.text = u" */ ;"
                mutationAfter = "----> after: " + node.getParent().getText().split(u'/*')[0]

                if self.verbose:
                    print(mutationAfter)
                mutatedTreesTexts.append((
                        "/* LittleDarwin generated mutant\nmutant type: nullifyReturnValue\n" + mutationBefore + "\n" + mutationAfter + "\n----> line number in original file: " + str(
                    node.getParent().start.line) + "\n*/ \n\n" + (
                            " ".join(tree.getText().rsplit("<EOF>", 1)))))
                # mutatedTreesTexts.append(tree.getText())
                node.symbol.text = copy.deepcopy(originalText0)
                node.getParent().getChild(2).symbol.text = copy.deepcopy(originalText2)

                if node.getParent().start.line in self.mutantsPerLine.keys():
                    self.mutantsPerLine[node.getParent().start.line] += 1
                else:
                    self.mutantsPerLine[node.getParent().start.line] = 1

            return mutatedTreesTexts

            # elif mode == "return_nodes":
            #     nodeList = list()
            #     expressionList = self.javaParseObject.seek(tree, JavaParser.ParExpressionContext)
            #
            #     for exp in expressionList:
            #         nodeList.append([exp, "negateConditionalsMutator"])

    def nullifyInputVariable(self, tree, mode="return_text", nodeIndex=None):
        assert isinstance(tree, JavaParser.CompilationUnitContext)

        if mode == "return_text":

            mutatedTreesTexts = list()
            methodDeclarationList = self.javaParseObject.seek(tree, JavaParser.MethodDeclarationContext)

            while len(methodDeclarationList) > 0:
                # tmpTree = copy.deepcopy(tree)
                methodDeclarationIndex = methodDeclarationList.pop()

                methodDeclaration = self.javaParseObject.getNode(tree, methodDeclarationIndex)
                assert isinstance(methodDeclaration, JavaParser.MethodDeclarationContext)

                try:
                    variableIDList = self.javaParseObject.seek(methodDeclaration.formalParameters(),
                                                               JavaParser.VariableDeclaratorIdContext)

                    if len(variableIDList) == 0:
                        continue  # no variables in this declaration

                    node = methodDeclaration.methodBody().block().getChild(0,
                                                                           TerminalNodeImpl)  # can fail on methods with no body

                except Exception as e:
                    continue

                mutationBefore = "----> before: " + node.getText()

                if self.verbose:
                    print(mutationBefore)

                originalText = copy.deepcopy(node.symbol.text)

                for var in variableIDList:
                    typeNode = self.javaParseObject.getNode(methodDeclaration, var)
                    assert isinstance(typeNode, JavaParser.VariableDeclaratorIdContext)

                    if typeNode.parentCtx.getChild(0, JavaParser.JTypeContext).getChild(0,
                                                                                        JavaParser.PrimitiveTypeContext) is not None:
                        continue  # primitive typed method

                    node.symbol.text = u'{ ' + typeNode.getText() + u' = null; '

                    mutationAfter = "----> after: " + node.getText()

                    if self.verbose:
                        print(mutationAfter)
                    mutatedTreesTexts.append((
                            "/* LittleDarwin generated mutant\nmutant type: nullifyInputVariable\n" + mutationBefore + "\n" + mutationAfter + "\n----> line number in original file: " + str(
                        node.parentCtx.start.line) + "\n----> mutated nodes: " + str(var) + "\n*/ \n\n" + (
                                " ".join(tree.getText().rsplit("<EOF>", 1)))))  # create compilable, readable code

                    node.symbol.text = copy.deepcopy(originalText)

                    if node.parentCtx.start.line in self.mutantsPerLine.keys():
                        self.mutantsPerLine[node.parentCtx.start.line] += 1
                    else:
                        self.mutantsPerLine[node.parentCtx.start.line] = 1

            return mutatedTreesTexts


        elif mode == "return_nodes":
            pass

        elif mode == "return_tree":
            pass

    def arithmeticOperatorReplacementBinary(self, tree, mode="return_text", nodeIndex=None):
        assert isinstance(tree, JavaParser.CompilationUnitContext)

        if mode == "return_text":

            mutatedTreesTexts = list()
            expressionList = self.javaParseObject.seek(tree, JavaParser.ExpressionContext)
            # nodeList = list()

            while len(expressionList) > 0:
                # tmpTree = copy.deepcopy(tree)
                expressionIndex = expressionList.pop()

                node = self.javaParseObject.getNode(tree, expressionIndex)
                assert isinstance(node, JavaParser.ExpressionContext)

                try:
                    if not (isinstance(node.children[0], JavaParser.ExpressionContext) and isinstance(node.children[1],
                                                                                                      TerminalNodeImpl) and isinstance(
                        node.children[2], JavaParser.ExpressionContext)):
                        continue  # not a binary expression
                except Exception as e:
                    continue

                if not (node.children[1].symbol.text == u"+" or node.children[1].symbol.text == u"-" or node.children[
                    1].symbol.text == u"*" or node.children[1].symbol.text == u"/" or node.children[
                            1].symbol.text == u"%"):
                    continue  # not an arithmetic operator

                if node.children[0].getText()[0] == '\"' or node.children[2].getText()[0] == '\"':
                    continue  # string concatenation, don't change

                # nodeList.append([expressionIndex, "ArithmeticOperatorReplacementBinary"])

                mutationBefore = "----> before: " + node.getText()

                if self.verbose:
                    print(mutationBefore)

                # if mode == "return_text" or mode == "return_tree":
                originalText = copy.deepcopy(node.children[1].symbol.text)

                if originalText == u"+":
                    node.children[1].symbol.text = u"-"
                elif originalText == u"-":
                    node.children[1].symbol.text = u"+"
                elif originalText == u"/":
                    node.children[1].symbol.text = u"*"
                elif originalText == u"*":
                    node.children[1].symbol.text = u"/"
                elif originalText == u"%":
                    node.children[1].symbol.text = u"/"
                else:
                    assert False

                mutationAfter = "----> after: " + node.getText()
                if self.verbose:
                    print(mutationAfter)
                mutatedTreesTexts.append((
                        "/* LittleDarwin generated mutant\nmutant type: arithmeticOperatorReplacementBinary\n" + mutationBefore + "\n" + mutationAfter + "\n----> line number in original file: " + str(
                    node.start.line) + "\n----> mutated nodes: " + str(expressionIndex) + "\n*/ \n\n" + (
                            " ".join(tree.getText().rsplit("<EOF>", 1)))))  # create compilable, readable code

                node.children[1].symbol.text = copy.deepcopy(originalText)

                if node.start.line in self.mutantsPerLine.keys():
                    self.mutantsPerLine[node.start.line] += 1
                else:
                    self.mutantsPerLine[node.start.line] = 1

            return mutatedTreesTexts


        elif mode == "return_nodes":
            # mutatedTreesTexts = list()
            expressionList = self.javaParseObject.seek(tree, JavaParser.ExpressionContext)
            nodeList = list()

            while len(expressionList) > 0:
                # tmpTree = copy.deepcopy(tree)
                expressionIndex = expressionList.pop()

                node = self.javaParseObject.getNode(tree, expressionIndex)
                assert isinstance(node, JavaParser.ExpressionContext)

                try:
                    if not (isinstance(node.children[0], JavaParser.ExpressionContext) and isinstance(node.children[1],
                                                                                                      TerminalNodeImpl) and isinstance(
                        node.children[2], JavaParser.ExpressionContext)):
                        continue  # not a binary expression
                except Exception as e:
                    continue

                if not (node.children[1].symbol.text == u"+" or node.children[1].symbol.text == u"-" or node.children[
                    1].symbol.text == u"*" or node.children[1].symbol.text == u"/" or node.children[
                            1].symbol.text == u"%"):
                    continue  # not an arithmetic operator

                if node.children[0].getText()[0] == '\"' or node.children[2].getText()[0] == '\"':
                    continue  # string concatenation, don't change

                nodeList.append([expressionIndex, "ArithmeticOperatorReplacementBinary"])

            return nodeList

        elif mode == "return_tree":
            assert nodeIndex is not None

            node = self.javaParseObject.getNode(tree, nodeIndex)
            assert isinstance(node, JavaParser.ExpressionContext)

            try:
                if not (isinstance(node.children[0], JavaParser.ExpressionContext) and isinstance(node.children[1],
                                                                                                  TerminalNodeImpl) and isinstance(
                    node.children[2], JavaParser.ExpressionContext)):
                    assert False

            except Exception as e:
                assert False

            if not (node.children[1].symbol.text == u"+" or node.children[1].symbol.text == u"-" or node.children[
                1].symbol.text == u"*" or node.children[1].symbol.text == u"/" or node.children[1].symbol.text == u"%"):
                assert False

            if node.children[0].getText()[0] == '\"' or node.children[2].getText()[0] == '\"':
                assert False

            originalText = node.children[1].symbol.text

            if originalText == u"+":
                node.children[1].symbol.oldText = u"+"
                node.children[1].symbol.text = u"-"
            elif originalText == u"-":
                node.children[1].symbol.oldText = u"-"
                node.children[1].symbol.text = u"+"
            elif originalText == u"/":
                node.children[1].symbol.oldText = u"/"
                node.children[1].symbol.text = u"*"
            elif originalText == u"*":
                node.children[1].symbol.oldText = u"*"
                node.children[1].symbol.text = u"/"
            elif originalText == u"%":
                node.children[1].symbol.oldText = u"%"
                node.children[1].symbol.text = u"/"
            else:
                assert False

            return tree

    def arithmeticOperatorReplacementUnary(self, tree, mode="return_text", nodeIndex=None):
        assert isinstance(tree, JavaParser.CompilationUnitContext)

        if mode == "return_text":
            expressionList = self.javaParseObject.seek(tree, JavaParser.ExpressionContext)
            mutatedTreesTexts = list()
            # nodeList = list()

            while len(expressionList) > 0:
                # tmpTree = copy.deepcopy(tree)
                expressionIndex = expressionList.pop()

                node = self.javaParseObject.getNode(tree, expressionIndex)
                assert isinstance(node, JavaParser.ExpressionContext)

                try:
                    if not (isinstance(node.children[0], TerminalNodeImpl) and isinstance(node.children[1],
                                                                                          JavaParser.ExpressionContext)):
                        continue  # not a unary expression
                except Exception as e:
                    continue

                if not (node.children[0].symbol.text == u"+" or node.children[0].symbol.text == u"-"):
                    continue  # not an arithmetic operator

                mutationBefore = "----> before: " + node.getText()
                if self.verbose:
                    print(mutationBefore)
                originalText = copy.deepcopy(node.children[0].symbol.text)

                if originalText == u"+":
                    node.children[0].symbol.text = u"-"
                elif originalText == u"-":
                    node.children[0].symbol.text = u"+"
                else:
                    assert False

                mutationAfter = "----> after: " + node.getText()
                if self.verbose:
                    print(mutationAfter)
                mutatedTreesTexts.append((
                        "/* LittleDarwin generated mutant\nmutant type: arithmeticOperatorReplacementUnary\n" + mutationBefore + "\n" + mutationAfter + "\n----> line number in original file: " + str(
                    node.start.line) + "\n----> mutated nodes: " + str(expressionIndex) + "\n*/ \n\n" + (
                            " ".join(tree.getText().rsplit("<EOF>", 1)))))  # create compilable, readable code

                node.children[0].symbol.text = copy.deepcopy(originalText)

                if node.start.line in self.mutantsPerLine.keys():
                    self.mutantsPerLine[node.start.line] += 1
                else:
                    self.mutantsPerLine[node.start.line] = 1

            return mutatedTreesTexts

        elif mode == "return_nodes":
            # mutatedTreesTexts = list()
            expressionList = self.javaParseObject.seek(tree, JavaParser.ExpressionContext)
            nodeList = list()

            while len(expressionList) > 0:
                # tmpTree = copy.deepcopy(tree)
                expressionIndex = expressionList.pop()

                node = self.javaParseObject.getNode(tree, expressionIndex)
                assert isinstance(node, JavaParser.ExpressionContext)

                try:
                    if not (isinstance(node.children[0], TerminalNodeImpl) and isinstance(node.children[1],
                                                                                          JavaParser.ExpressionContext)):
                        continue  # not a unary expression
                except Exception as e:
                    continue

                if not (node.children[0].symbol.text == u"+" or node.children[0].symbol.text == u"-"):
                    continue  # not an arithmetic operator

                nodeList.append([expressionIndex, "ArithmeticOperatorReplacementUnary"])

            return nodeList

        elif mode == "return_tree":
            assert nodeIndex is not None

            node = self.javaParseObject.getNode(tree, nodeIndex)
            assert isinstance(node, JavaParser.ExpressionContext)

            try:
                if not (isinstance(node.children[0], TerminalNodeImpl) and isinstance(node.children[1],
                                                                                      JavaParser.ExpressionContext)):
                    assert False

            except Exception as e:
                assert False

            if not (node.children[0].symbol.text == u"+" or node.children[0].symbol.text == u"-"):
                assert False

            originalText = node.children[0].symbol.text

            if originalText == u"+":
                node.children[0].symbol.oldText = u"+"
                node.children[0].symbol.text = u"-"
            elif originalText == u"-":
                node.children[0].symbol.oldText = u"-"
                node.children[0].symbol.text = u"+"
            else:
                assert False

            return tree

    def arithmeticOperatorReplacementShortcut(self, tree, mode="return_text", nodeIndex=None):
        assert isinstance(tree, JavaParser.CompilationUnitContext)

        if mode == "return_text":

            mutatedTreesTexts = list()
            expressionList = self.javaParseObject.seek(tree, JavaParser.ExpressionContext)

            while len(expressionList) > 0:
                # tmpTree = copy.deepcopy(tree)
                expressionIndex = expressionList.pop()

                node = self.javaParseObject.getNode(tree, expressionIndex)
                assert isinstance(node, JavaParser.ExpressionContext)

                try:
                    if isinstance(node.children[0], TerminalNodeImpl) and isinstance(node.children[1],
                                                                                     JavaParser.ExpressionContext):
                        terminalChild = 0
                    elif isinstance(node.children[1], TerminalNodeImpl) and isinstance(node.children[0],
                                                                                       JavaParser.ExpressionContext):
                        terminalChild = 1
                    else:
                        continue  # not a shortcut expression
                except Exception as e:
                    continue

                if not (node.children[terminalChild].symbol.text == u"++" or node.children[
                    terminalChild].symbol.text == u"--"):
                    continue  # not an arithmetic operator

                mutationBefore = "----> before: " + node.getText()
                if self.verbose:
                    print(mutationBefore)
                originalText = copy.deepcopy(node.children[terminalChild].symbol.text)

                if originalText == u"++":
                    node.children[terminalChild].symbol.text = u"--"
                elif originalText == u"--":
                    node.children[terminalChild].symbol.text = u"++"
                else:
                    assert False

                mutationAfter = "----> after: " + node.getText()
                if self.verbose:
                    print(mutationAfter)

                mutatedTreesTexts.append((
                        "/* LittleDarwin generated mutant\nmutant type: arithmeticOperatorReplacementShortcut\n" + mutationBefore + "\n" + mutationAfter + "\n----> line number in original file: " + str(
                    node.start.line) + "\n----> mutated nodes: " + str(expressionIndex) + "\n*/ \n\n" + (
                            " ".join(tree.getText().rsplit("<EOF>", 1)))))  # create compilable, readable code

                node.children[terminalChild].symbol.text = copy.deepcopy(originalText)

                if node.start.line in self.mutantsPerLine.keys():
                    self.mutantsPerLine[node.start.line] += 1
                else:
                    self.mutantsPerLine[node.start.line] = 1

            return mutatedTreesTexts

        elif mode == "return_nodes":
            nodeList = list()
            expressionList = self.javaParseObject.seek(tree, JavaParser.ExpressionContext)

            while len(expressionList) > 0:
                # tmpTree = copy.deepcopy(tree)
                expressionIndex = expressionList.pop()

                node = self.javaParseObject.getNode(tree, expressionIndex)
                assert isinstance(node, JavaParser.ExpressionContext)

                try:
                    if isinstance(node.children[0], TerminalNodeImpl) and isinstance(node.children[1],
                                                                                     JavaParser.ExpressionContext):
                        terminalChild = 0
                    elif isinstance(node.children[1], TerminalNodeImpl) and isinstance(node.children[0],
                                                                                       JavaParser.ExpressionContext):
                        terminalChild = 1
                    else:
                        continue  # not a shortcut expression
                except Exception as e:
                    continue

                if not (node.children[terminalChild].symbol.text == u"++" or node.children[
                    terminalChild].symbol.text == u"--"):
                    continue  # not an arithmetic operator

                nodeList.append([expressionIndex, "ArithmeticOperatorReplacementShortcut"])

            return nodeList

        elif mode == "return_tree":
            assert nodeIndex is not None

            node = self.javaParseObject.getNode(tree, nodeIndex)
            assert isinstance(node, JavaParser.ExpressionContext)

            try:

                if isinstance(node.children[0], TerminalNodeImpl) and isinstance(node.children[1],
                                                                                 JavaParser.ExpressionContext):
                    terminalChild = 0
                elif isinstance(node.children[1], TerminalNodeImpl) and isinstance(node.children[0],
                                                                                   JavaParser.ExpressionContext):
                    terminalChild = 1
                else:
                    assert False
            except Exception as e:
                assert False

            if not (node.children[terminalChild].symbol.text == u"++" or node.children[
                terminalChild].symbol.text == u"--"):
                assert False

            originalText = node.children[terminalChild].symbol.text

            if originalText == u"++":
                node.children[terminalChild].symbol.oldText = u"++"
                node.children[terminalChild].symbol.text = u"--"
            elif originalText == u"--":
                node.children[terminalChild].symbol.oldText = u"--"
                node.children[terminalChild].symbol.text = u"++"
            else:
                assert False

            return tree

    # def arithmeticOperatorInsertionUnary(self, tree):
    #     pass
    #
    # def arithmeticOperatorInsertionShortcut(self, tree):
    #     pass
    #
    # def arithmeticOperatorDeletionUnary(self, tree):
    #     pass
    #
    # def arithmeticOperatorDeletionShortcut(self, tree):
    #     pass

    def relationalOperatorReplacement(self, tree, mode="return_text",
                                      nodeIndex=None):  # executionCount by negateConditionals
        assert isinstance(tree, JavaParser.CompilationUnitContext)

        if mode == "return_text":
            mutatedTreesTexts = list()
            expressionList = self.javaParseObject.seek(tree, JavaParser.ExpressionContext)

            while len(expressionList) > 0:
                # tmpTree = copy.deepcopy(tree)
                expressionIndex = expressionList.pop()

                node = self.javaParseObject.getNode(tree, expressionIndex)
                assert isinstance(node, JavaParser.ExpressionContext)

                try:
                    if not (isinstance(node.children[0], JavaParser.ExpressionContext) and isinstance(node.children[1],
                                                                                                      TerminalNodeImpl) and isinstance(
                        node.children[2], JavaParser.ExpressionContext)):
                        continue  # not a binary expression
                except Exception as e:
                    continue

                if not (node.children[1].symbol.text == u">" or node.children[1].symbol.text == u"<" or node.children[
                    1].symbol.text == u">=" or node.children[1].symbol.text == u"<=" or node.children[
                            1].symbol.text == u"==" or node.children[1].symbol.text == u"!="):
                    continue  # not a relation operator

                mutationBefore = "----> before: " + node.getText()
                if self.verbose:
                    print(mutationBefore)

                originalText = copy.deepcopy(node.children[1].symbol.text)

                if originalText == u">":
                    node.children[1].symbol.text = u"<="
                elif originalText == u"<":
                    node.children[1].symbol.text = u">="
                elif originalText == u"<=":
                    node.children[1].symbol.text = u">"
                elif originalText == u">=":
                    node.children[1].symbol.text = u"<"
                elif originalText == u"!=":
                    node.children[1].symbol.text = u"=="
                elif originalText == u"==":
                    node.children[1].symbol.text = u"!="
                else:
                    assert False

                mutationAfter = "----> after: " + node.getText()

                if self.verbose:
                    print(mutationAfter)

                mutatedTreesTexts.append((
                    (
                            "/* LittleDarwin generated mutant\nmutant type: relationalOperatorReplacement\n" + mutationBefore + "\n" + mutationAfter + "\n----> line number in original file: " + str(
                        node.start.line) + "\n----> mutated nodes: " + str(expressionIndex) + "\n*/ \n\n" + (
                                " ".join(tree.getText().rsplit("<EOF>", 1))))))  # create compilable, readable code

                node.children[1].symbol.text = copy.deepcopy(originalText)

                if node.start.line in self.mutantsPerLine.keys():
                    self.mutantsPerLine[node.start.line] += 1
                else:
                    self.mutantsPerLine[node.start.line] = 1

            return mutatedTreesTexts

        elif mode == "return_nodes":
            nodeList = list()
            expressionList = self.javaParseObject.seek(tree, JavaParser.ExpressionContext)

            while len(expressionList) > 0:
                # tmpTree = copy.deepcopy(tree)
                expressionIndex = expressionList.pop()

                node = self.javaParseObject.getNode(tree, expressionIndex)
                assert isinstance(node, JavaParser.ExpressionContext)

                try:
                    if not (isinstance(node.children[0], JavaParser.ExpressionContext) and isinstance(node.children[1],
                                                                                                      TerminalNodeImpl) and isinstance(
                        node.children[2], JavaParser.ExpressionContext)):
                        continue  # not a binary expression
                except Exception as e:
                    continue

                if not (node.children[1].symbol.text == u">" or node.children[1].symbol.text == u"<" or node.children[
                    1].symbol.text == u">=" or node.children[1].symbol.text == u"<=" or node.children[
                            1].symbol.text == u"==" or node.children[1].symbol.text == u"!="):
                    continue  # not a relation operator

                nodeList.append([expressionIndex, "RelationalOperatorReplacement"])

            return nodeList

        elif mode == "return_tree":
            assert nodeIndex is not None

            node = self.javaParseObject.getNode(tree, nodeIndex)
            assert isinstance(node, JavaParser.ExpressionContext)

            try:
                if not (isinstance(node.children[0], JavaParser.ExpressionContext) and isinstance(node.children[1],
                                                                                                  TerminalNodeImpl) and isinstance(
                    node.children[2], JavaParser.ExpressionContext)):
                    assert False
            except Exception as e:
                assert False

            if not (node.children[1].symbol.text == u">" or node.children[1].symbol.text == u"<" or node.children[
                1].symbol.text == u">=" or node.children[1].symbol.text == u"<=" or node.children[
                        1].symbol.text == u"==" or node.children[1].symbol.text == u"!="):
                assert False

            originalText = node.children[1].symbol.text

            if originalText == u">":
                node.children[1].symbol.oldText = u">"
                node.children[1].symbol.text = u"<="
            elif originalText == u"<":
                node.children[1].symbol.oldText = u"<"
                node.children[1].symbol.text = u">="
            elif originalText == u"<=":
                node.children[1].symbol.oldText = u"<="
                node.children[1].symbol.text = u">"
            elif originalText == u">=":
                node.children[1].symbol.oldText = u">="
                node.children[1].symbol.text = u"<"
            elif originalText == u"!=":
                node.children[1].symbol.oldText = u"!="
                node.children[1].symbol.text = u"=="
            elif originalText == u"==":
                node.children[1].symbol.oldText = u"=="
                node.children[1].symbol.text = u"!="
            else:
                assert False

            return tree

    def conditionalOperatorReplacement(self, tree, mode="return_text", nodeIndex=None):
        assert isinstance(tree, JavaParser.CompilationUnitContext)

        if mode == "return_text":

            mutatedTreesTexts = list()
            expressionList = self.javaParseObject.seek(tree, JavaParser.ExpressionContext)

            while len(expressionList) > 0:
                # tmpTree = copy.deepcopy(tree)
                expressionIndex = expressionList.pop()

                node = self.javaParseObject.getNode(tree, expressionIndex)
                assert isinstance(node, JavaParser.ExpressionContext)

                try:
                    if not (isinstance(node.children[0], JavaParser.ExpressionContext) and isinstance(node.children[1],
                                                                                                      TerminalNodeImpl) and isinstance(
                        node.children[2], JavaParser.ExpressionContext)):
                        continue  # not a binary expression
                except Exception as e:
                    continue

                if not (node.children[1].symbol.text == u"&&" or node.children[1].symbol.text == u"||"):
                    continue  # not a conditional operator (non-lazy ones executionCount in logical operators)

                mutationBefore = "----> before: " + node.getText()
                if self.verbose:
                    print(mutationBefore)

                originalText = copy.deepcopy(node.children[1].symbol.text)

                if originalText == u"&&":
                    node.children[1].symbol.text = u"||"
                elif originalText == u"||":
                    node.children[1].symbol.text = u"&&"
                else:
                    assert False

                mutationAfter = "----> after: " + node.getText()
                if self.verbose:
                    print(mutationAfter)
                mutatedTreesTexts.append((
                        "/* LittleDarwin generated mutant\nmutant type: conditionalOperatorReplacement\n" + mutationBefore + "\n" + mutationAfter + "\n----> line number in original file: " + str(
                    node.start.line) + "\n----> mutated nodes: " + str(expressionIndex) + "\n*/ \n\n" + (
                            " ".join(tree.getText().rsplit("<EOF>", 1)))))  # create compilable, readable code

                node.children[1].symbol.text = copy.deepcopy(originalText)

                if node.start.line in self.mutantsPerLine.keys():
                    self.mutantsPerLine[node.start.line] += 1
                else:
                    self.mutantsPerLine[node.start.line] = 1


            return mutatedTreesTexts

        elif mode == "return_nodes":

            nodeList = list()
            expressionList = self.javaParseObject.seek(tree, JavaParser.ExpressionContext)

            while len(expressionList) > 0:
                # tmpTree = copy.deepcopy(tree)
                expressionIndex = expressionList.pop()

                node = self.javaParseObject.getNode(tree, expressionIndex)
                assert isinstance(node, JavaParser.ExpressionContext)

                try:
                    if not (isinstance(node.children[0], JavaParser.ExpressionContext) and isinstance(node.children[1],
                                                                                                      TerminalNodeImpl) and isinstance(
                        node.children[2], JavaParser.ExpressionContext)):
                        continue  # not a binary expression
                except Exception as e:
                    continue

                if not (node.children[1].symbol.text == u"&&" or node.children[1].symbol.text == u"||"):
                    continue  # not a conditional operator (non-lazy ones executionCount in logical operators)

                nodeList.append([expressionIndex, "ConditionalOperatorReplacement"])

            return nodeList

        elif mode == "return_tree":
            assert nodeIndex is not None

            node = self.javaParseObject.getNode(tree, nodeIndex)
            assert isinstance(node, JavaParser.ExpressionContext)

            try:
                if not (isinstance(node.children[0], JavaParser.ExpressionContext) and isinstance(node.children[1],
                                                                                                  TerminalNodeImpl) and isinstance(
                    node.children[2], JavaParser.ExpressionContext)):
                    assert False
            except Exception as e:
                assert False

            if not (node.children[1].symbol.text == u"&&" or node.children[1].symbol.text == u"||"):
                assert False

            originalText = node.children[1].symbol.text

            if originalText == u"&&":
                node.children[1].symbol.oldText = u"&&"
                node.children[1].symbol.text = u"||"
            elif originalText == u"||":
                node.children[1].symbol.oldText = u"||"
                node.children[1].symbol.text = u"&&"
            else:
                assert False

            return tree

    def conditionalOperatorDeletion(self, tree, mode="return_text",
                                    nodeIndex=None):  # executionCount by negateConditionals
        assert isinstance(tree, JavaParser.CompilationUnitContext)

        if mode == "return_text":

            mutatedTreesTexts = list()
            expressionList = self.javaParseObject.seek(tree, JavaParser.ExpressionContext)

            while len(expressionList) > 0:
                # tmpTree = copy.deepcopy(tree)
                expressionIndex = expressionList.pop()

                node = self.javaParseObject.getNode(tree, expressionIndex)
                assert isinstance(node, JavaParser.ExpressionContext)

                try:
                    if not (isinstance(node.children[0], TerminalNodeImpl) and isinstance(node.children[1],
                                                                                          JavaParser.ExpressionContext)):
                        continue  # not a unary expression
                except Exception as e:
                    continue

                if not (node.children[0].symbol.text == u"!"):
                    continue  # not a unary conditional operator

                mutationBefore = "----> before: " + node.getText()
                if self.verbose:
                    print(mutationBefore)
                originalText = copy.deepcopy(node.children[0].symbol.text)

                if originalText == u"!":
                    node.children[0].symbol.text = u" "
                else:
                    assert False

                mutationAfter = "----> after: " + node.getText()
                if self.verbose:
                    print(mutationAfter)
                mutatedTreesTexts.append((
                        "/* LittleDarwin generated mutant\nmutant type: conditionalOperatorDeletion\n" + mutationBefore + "\n" + mutationAfter + "\n----> line number in original file: " + str(
                    node.start.line) + "\n----> mutated nodes: " + str(expressionIndex) + "\n*/ \n\n" + (
                            " ".join(tree.getText().rsplit("<EOF>", 1)))))  # create compilable, readable code

                node.children[0].symbol.text = copy.deepcopy(originalText)

                if node.start.line in self.mutantsPerLine.keys():
                    self.mutantsPerLine[node.start.line] += 1
                else:
                    self.mutantsPerLine[node.start.line] = 1

            return mutatedTreesTexts

        elif mode == "return_nodes":

            nodeList = list()
            expressionList = self.javaParseObject.seek(tree, JavaParser.ExpressionContext)

            while len(expressionList) > 0:
                # tmpTree = copy.deepcopy(tree)
                expressionIndex = expressionList.pop()

                node = self.javaParseObject.getNode(tree, expressionIndex)
                assert isinstance(node, JavaParser.ExpressionContext)

                try:
                    if not (isinstance(node.children[0], TerminalNodeImpl) and isinstance(node.children[1],
                                                                                          JavaParser.ExpressionContext)):
                        continue  # not a unary expression
                except Exception as e:
                    continue

                if not (node.children[0].symbol.text == u"!"):
                    continue  # not a unary conditional operator

                nodeList.append([expressionIndex, "ConditionalOperatorDeletion"])

            return nodeList

        elif mode == "return_tree":
            assert nodeIndex is not None
            node = self.javaParseObject.getNode(tree, nodeIndex)
            assert isinstance(node, JavaParser.ExpressionContext)

            try:
                if not (isinstance(node.children[0], TerminalNodeImpl) and isinstance(node.children[1],
                                                                                      JavaParser.ExpressionContext)):
                    assert False
            except Exception as e:
                assert False

            if not (node.children[0].symbol.text == u"!"):
                assert False

            originalText = node.children[0].symbol.text

            if originalText == u"!":
                node.children[0].symbol.oldText = u"!"
                node.children[0].symbol.text = u" "
            else:
                assert False

            return tree

    # def conditionalOperatorInsertion(self, tree):   # executionCount by negateConditionals, both generate too many mutations
    #     pass

    def shiftOperatorReplacement(self, tree, mode="return_text", nodeIndex=None):
        assert isinstance(tree, JavaParser.CompilationUnitContext)

        if mode == "return_text":

            mutatedTreesTexts = list()
            expressionList = self.javaParseObject.seek(tree, JavaParser.ExpressionContext)

            while len(expressionList) > 0:
                # tmpTree = copy.deepcopy(tree)
                expressionIndex = expressionList.pop()

                node = self.javaParseObject.getNode(tree, expressionIndex)
                assert isinstance(node, JavaParser.ExpressionContext)

                try:
                    if (isinstance(node.children[0], JavaParser.ExpressionContext) and isinstance(node.children[1],
                                                                                                  TerminalNodeImpl) and isinstance(
                        node.children[2], TerminalNodeImpl) and isinstance(node.children[3],
                                                                           JavaParser.ExpressionContext)):
                        threeTerminals = False
                    elif (isinstance(node.children[0], JavaParser.ExpressionContext) and isinstance(node.children[1],
                                                                                                    TerminalNodeImpl) and isinstance(
                        node.children[2], TerminalNodeImpl) and isinstance(node.children[3],
                                                                           TerminalNodeImpl) and isinstance(
                        node.children[4], JavaParser.ExpressionContext)):
                        threeTerminals = True
                    else:
                        continue  # not a binary shift expression
                except Exception as e:
                    continue

                try:
                    if (threeTerminals is False) and (
                            (node.children[1].symbol.text == u"<" and node.children[2].symbol.text == u"<") or (
                            node.children[1].symbol.text == u">" and node.children[2].symbol.text == u">")):
                        pass

                    elif (threeTerminals is True) and (
                            node.children[1].symbol.text == u">" and node.children[2].symbol.text == u">" and
                            node.children[3].symbol.text == u">"):
                        pass

                    else:
                        continue  # not a shift operator

                except Exception as e:
                    continue

                mutationBefore = "----> before: " + node.getText()
                if self.verbose:
                    print(mutationBefore)

                if threeTerminals:
                    originalText1 = copy.deepcopy(node.children[1].symbol.text)
                    originalText2 = copy.deepcopy(node.children[2].symbol.text)
                    originalText3 = copy.deepcopy(node.children[3].symbol.text)

                    if originalText1 == u">" and originalText2 == u">" and originalText3 == u">":
                        node.children[3].symbol.text = u" "
                    else:
                        assert False

                else:
                    originalText1 = copy.deepcopy(node.children[1].symbol.text)
                    originalText2 = copy.deepcopy(node.children[2].symbol.text)

                    if originalText1 == u">" and originalText2 == u">":
                        node.children[1].symbol.text = u"<"
                        node.children[2].symbol.text = u"<"
                    elif originalText1 == u"<" and originalText2 == u"<":
                        node.children[1].symbol.text = u">"
                        node.children[2].symbol.text = u">"
                    else:
                        assert False

                mutationAfter = "----> after: " + node.getText()
                if self.verbose:
                    print(mutationAfter)

                mutatedTreesTexts.append((
                        "/* LittleDarwin generated mutant\nmutant type: shiftOperatorReplacement\n" + mutationBefore + "\n" + mutationAfter + "\n----> line number in original file: " + str(
                    node.start.line) + "\n----> mutated nodes: " + str(expressionIndex) + "\n*/ \n\n" + (
                            " ".join(tree.getText().rsplit("<EOF>", 1)))))  # create compilable, readable code

                node.children[1].symbol.text = copy.deepcopy(originalText1)
                node.children[2].symbol.text = copy.deepcopy(originalText2)
                if threeTerminals:
                    node.children[3].symbol.text = copy.deepcopy(originalText3)

                if node.start.line in self.mutantsPerLine.keys():
                    self.mutantsPerLine[node.start.line] += 1
                else:
                    self.mutantsPerLine[node.start.line] = 1

            return mutatedTreesTexts

        elif mode == "return_nodes":

            nodeList = list()
            expressionList = self.javaParseObject.seek(tree, JavaParser.ExpressionContext)

            while len(expressionList) > 0:
                # tmpTree = copy.deepcopy(tree)
                expressionIndex = expressionList.pop()

                node = self.javaParseObject.getNode(tree, expressionIndex)
                assert isinstance(node, JavaParser.ExpressionContext)

                try:
                    if (isinstance(node.children[0], JavaParser.ExpressionContext) and isinstance(node.children[1],
                                                                                                  TerminalNodeImpl) and isinstance(
                        node.children[2], TerminalNodeImpl) and isinstance(node.children[3],
                                                                           JavaParser.ExpressionContext)):
                        threeTerminals = False
                    elif (isinstance(node.children[0], JavaParser.ExpressionContext) and isinstance(node.children[1],
                                                                                                    TerminalNodeImpl) and isinstance(
                        node.children[2], TerminalNodeImpl) and isinstance(node.children[3],
                                                                           TerminalNodeImpl) and isinstance(
                        node.children[4], JavaParser.ExpressionContext)):
                        threeTerminals = True
                    else:
                        continue  # not a binary shift expression
                except Exception as e:
                    continue

                try:
                    if (threeTerminals is False) and (
                            (node.children[1].symbol.text == u"<" and node.children[2].symbol.text == u"<") or (
                            node.children[1].symbol.text == u">" and node.children[2].symbol.text == u">")):
                        pass

                    elif (threeTerminals is True) and (
                            node.children[1].symbol.text == u">" and node.children[2].symbol.text == u">" and
                            node.children[3].symbol.text == u">"):
                        pass

                    else:
                        continue  # not a shift operator

                except Exception as e:
                    continue

                nodeList.append([expressionIndex, "ShiftOperatorReplacement"])

            return nodeList

        if mode == "return_tree":
            assert nodeIndex is not None

            node = self.javaParseObject.getNode(tree, nodeIndex)
            assert isinstance(node, JavaParser.ExpressionContext)

            try:
                if (isinstance(node.children[0], JavaParser.ExpressionContext) and isinstance(node.children[1],
                                                                                              TerminalNodeImpl) and isinstance(
                    node.children[2], TerminalNodeImpl) and isinstance(node.children[3],
                                                                       JavaParser.ExpressionContext)):
                    threeTerminals = False
                elif (isinstance(node.children[0], JavaParser.ExpressionContext) and isinstance(node.children[1],
                                                                                                TerminalNodeImpl) and isinstance(
                    node.children[2], TerminalNodeImpl) and isinstance(node.children[3],
                                                                       TerminalNodeImpl) and isinstance(
                    node.children[4], JavaParser.ExpressionContext)):
                    threeTerminals = True

                else:
                    assert False
            except Exception as e:
                assert False

            try:
                if (threeTerminals is False) and (
                        (node.children[1].symbol.text == u"<" and node.children[2].symbol.text == u"<") or (
                        node.children[1].symbol.text == u">" and node.children[2].symbol.text == u">")):
                    pass

                elif (threeTerminals is True) and (
                        node.children[1].symbol.text == u">" and node.children[2].symbol.text == u">" and
                        node.children[3].symbol.text == u">"):
                    pass

                else:
                    assert False

            except Exception as e:
                assert False

            if threeTerminals:
                originalText1 = node.children[1].symbol.text
                originalText2 = node.children[2].symbol.text
                originalText3 = node.children[3].symbol.text

                if originalText1 == u">" and originalText2 == u">" and originalText3 == u">":
                    node.children[3].symbol.oldText = u">"
                    node.children[3].symbol.text = u" "
                else:
                    assert False

            else:
                originalText1 = node.children[1].symbol.text
                originalText2 = node.children[2].symbol.text

                if originalText1 == u">" and originalText2 == u">":
                    node.children[1].symbol.oldText = u">"
                    node.children[2].symbol.oldText = u">"
                    node.children[1].symbol.text = u"<"
                    node.children[2].symbol.text = u"<"
                elif originalText1 == u"<" and originalText2 == u"<":
                    node.children[1].symbol.oldText = u"<"
                    node.children[2].symbol.oldText = u"<"
                    node.children[1].symbol.text = u">"
                    node.children[2].symbol.text = u">"
                else:
                    assert False

            return tree

    def logicalOperatorReplacement(self, tree, mode="return_text", nodeIndex=None):
        assert isinstance(tree, JavaParser.CompilationUnitContext)

        if mode == "return_text":
            mutatedTreesTexts = list()
            expressionList = self.javaParseObject.seek(tree, JavaParser.ExpressionContext)

            while len(expressionList) > 0:
                # tmpTree = copy.deepcopy(tree)
                expressionIndex = expressionList.pop()

                node = self.javaParseObject.getNode(tree, expressionIndex)
                assert isinstance(node, JavaParser.ExpressionContext)

                try:
                    if not (isinstance(node.children[0], JavaParser.ExpressionContext) and isinstance(node.children[1],
                                                                                                      TerminalNodeImpl) and isinstance(
                        node.children[2], JavaParser.ExpressionContext)):
                        continue  # not a binary expression
                except Exception as e:
                    continue

                if not (node.children[1].symbol.text == u"&" or node.children[1].symbol.text == u"|" or node.children[
                    1].symbol.text == u"^"):
                    continue  # not a logical operator

                mutationBefore = "----> before: " + node.getText()
                if self.verbose:
                    print(mutationBefore)

                originalText = copy.deepcopy(node.children[1].symbol.text)

                if originalText == u"&":
                    node.children[1].symbol.text = u"|"
                elif originalText == u"|":
                    node.children[1].symbol.text = u"^"
                elif originalText == u"^":
                    node.children[1].symbol.text = u"&"
                else:
                    assert False

                mutationAfter = "----> after: " + node.getText()
                if self.verbose:
                    print(mutationAfter)
                mutatedTreesTexts.append((
                        "/* LittleDarwin generated mutant\nmutant type: logicalOperatorReplacement\n" + mutationBefore + "\n" + mutationAfter + "\n----> line number in original file: " + str(
                    node.start.line) + "\n----> mutated nodes: " + str(expressionIndex) + "\n*/ \n\n" + (
                            " ".join(tree.getText().rsplit("<EOF>", 1)))))  # create compilable, readable code

                node.children[1].symbol.text = copy.deepcopy(originalText)

                if node.start.line in self.mutantsPerLine.keys():
                    self.mutantsPerLine[node.start.line] += 1
                else:
                    self.mutantsPerLine[node.start.line] = 1

            return mutatedTreesTexts

        elif mode == "return_nodes":
            nodeList = list()
            expressionList = self.javaParseObject.seek(tree, JavaParser.ExpressionContext)

            while len(expressionList) > 0:
                # tmpTree = copy.deepcopy(tree)
                expressionIndex = expressionList.pop()

                node = self.javaParseObject.getNode(tree, expressionIndex)
                assert isinstance(node, JavaParser.ExpressionContext)

                try:
                    if not (isinstance(node.children[0], JavaParser.ExpressionContext) and isinstance(node.children[1],
                                                                                                      TerminalNodeImpl) and isinstance(
                        node.children[2], JavaParser.ExpressionContext)):
                        continue  # not a binary expression
                except Exception as e:
                    continue

                if not (node.children[1].symbol.text == u"&" or node.children[1].symbol.text == u"|" or node.children[
                    1].symbol.text == u"^"):
                    continue  # not a logical operator
                nodeList.append([expressionIndex, "LogicalOperatorReplacement"])

            return nodeList

        elif mode == "return_tree":
            assert nodeIndex is not None

            node = self.javaParseObject.getNode(tree, nodeIndex)
            assert isinstance(node, JavaParser.ExpressionContext)

            try:
                if not (isinstance(node.children[0], JavaParser.ExpressionContext) and isinstance(node.children[1],
                                                                                                  TerminalNodeImpl) and isinstance(
                    node.children[2], JavaParser.ExpressionContext)):
                    assert False
            except Exception as e:
                assert False

            if not (node.children[1].symbol.text == u"&" or node.children[1].symbol.text == u"|" or node.children[
                1].symbol.text == u"^"):
                assert False

            originalText = node.children[1].symbol.text

            if originalText == u"&":
                node.children[1].symbol.oldText = u"&"
                node.children[1].symbol.text = u"|"
            elif originalText == u"|":
                node.children[1].symbol.oldText = u"|"
                node.children[1].symbol.text = u"^"
            elif originalText == u"^":
                node.children[1].symbol.oldText = u"^"
                node.children[1].symbol.text = u"&"
            else:
                assert False

            return tree

    # def logicalOperatorInsertion(self, tree):
    #     pass
    #
    # def logicalOperatorDeletion(self, tree):
    #     pass

    def assignmentOperatorReplacementShortcut(self, tree, mode="return_text", nodeIndex=None):
        assert isinstance(tree, JavaParser.CompilationUnitContext)

        if mode == "return_text":
            mutatedTreesTexts = list()
            expressionList = self.javaParseObject.seek(tree, JavaParser.ExpressionContext)

            while len(expressionList) > 0:
                # tmpTree = copy.deepcopy(tree)
                expressionIndex = expressionList.pop()

                node = self.javaParseObject.getNode(tree, expressionIndex)
                assert isinstance(node, JavaParser.ExpressionContext)

                try:
                    if not (isinstance(node.children[0], JavaParser.ExpressionContext) and isinstance(node.children[1],
                                                                                                      TerminalNodeImpl) and isinstance(
                        node.children[2], JavaParser.ExpressionContext)):
                        continue  # not a binary expression
                except Exception as e:
                    continue

                if not (node.children[1].symbol.text == u"+=" or node.children[1].symbol.text == u"-=" or node.children[
                    1].symbol.text == u"*=" or node.children[1].symbol.text == u"/=" or node.children[
                            1].symbol.text == u"%=" or node.children[1].symbol.text == u"&=" or node.children[
                            1].symbol.text == u"|=" or node.children[1].symbol.text == u"^=" or node.children[
                            1].symbol.text == u"<<=" or node.children[1].symbol.text == u">>=" or node.children[
                            1].symbol.text == u">>>="):
                    continue  # not an assignment operator

                mutationBefore = "----> before: " + node.getText()
                if self.verbose:
                    print(mutationBefore)

                originalText = copy.deepcopy(node.children[1].symbol.text)

                if originalText == u"+=":
                    node.children[1].symbol.text = u"-="
                elif originalText == u"-=":
                    node.children[1].symbol.text = u"+="
                elif originalText == u"/=":
                    node.children[1].symbol.text = u"*="
                elif originalText == u"*=":
                    node.children[1].symbol.text = u"/="
                elif originalText == u"%=":
                    node.children[1].symbol.text = u"/="
                elif originalText == u"&=":
                    node.children[1].symbol.text = u"|="
                elif originalText == u"|=":
                    node.children[1].symbol.text = u"^="
                elif originalText == u"^=":
                    node.children[1].symbol.text = u"&="
                elif originalText == u">>=":
                    node.children[1].symbol.text = u">>>="
                elif originalText == u"<<=":
                    node.children[1].symbol.text = u">>="
                elif originalText == u">>>=":
                    node.children[1].symbol.text = u">>="
                else:
                    assert False

                mutationAfter = "----> after: " + node.getText()
                if self.verbose:
                    print(mutationAfter)
                mutatedTreesTexts.append((
                        "/* LittleDarwin generated mutant\nmutant type: assignmentOperatorReplacementShortcut\n" + mutationBefore + "\n" + mutationAfter + "\n----> line number in original file: " + str(
                    node.start.line) + "\n----> mutated nodes: " + str(expressionIndex) + "\n*/ \n\n" + (
                            " ".join(tree.getText().rsplit("<EOF>", 1)))))  # create compilable, readable code

                node.children[1].symbol.text = copy.deepcopy(originalText)

                if node.start.line in self.mutantsPerLine.keys():
                    self.mutantsPerLine[node.start.line] += 1
                else:
                    self.mutantsPerLine[node.start.line] = 1

            return mutatedTreesTexts

        elif mode == "return_nodes":
            nodeList = list()
            expressionList = self.javaParseObject.seek(tree, JavaParser.ExpressionContext)

            while len(expressionList) > 0:
                # tmpTree = copy.deepcopy(tree)
                expressionIndex = expressionList.pop()

                node = self.javaParseObject.getNode(tree, expressionIndex)
                assert isinstance(node, JavaParser.ExpressionContext)

                try:
                    if not (isinstance(node.children[0], JavaParser.ExpressionContext) and isinstance(node.children[1],
                                                                                                      TerminalNodeImpl) and isinstance(
                        node.children[2], JavaParser.ExpressionContext)):
                        continue  # not a binary expression
                except Exception as e:
                    continue

                if not (node.children[1].symbol.text == u"+=" or node.children[1].symbol.text == u"-=" or node.children[
                    1].symbol.text == u"*=" or node.children[1].symbol.text == u"/=" or node.children[
                            1].symbol.text == u"%=" or node.children[1].symbol.text == u"&=" or node.children[
                            1].symbol.text == u"|=" or node.children[1].symbol.text == u"^=" or node.children[
                            1].symbol.text == u"<<=" or node.children[1].symbol.text == u">>=" or node.children[
                            1].symbol.text == u">>>="):
                    continue  # not an assignment operator

                nodeList.append([expressionIndex, "AssignmentOperatorReplacementShortcut"])

            return nodeList

        elif mode == "return_tree":
            assert nodeIndex is not None
            node = self.javaParseObject.getNode(tree, nodeIndex)
            assert isinstance(node, JavaParser.ExpressionContext)

            try:
                if not (isinstance(node.children[0], JavaParser.ExpressionContext) and isinstance(node.children[1],
                                                                                                  TerminalNodeImpl) and isinstance(
                    node.children[2], JavaParser.ExpressionContext)):
                    assert False
            except Exception as e:
                assert False

            if not (node.children[1].symbol.text == u"+=" or node.children[1].symbol.text == u"-=" or node.children[
                1].symbol.text == u"*=" or node.children[1].symbol.text == u"/=" or node.children[
                        1].symbol.text == u"%=" or node.children[1].symbol.text == u"&=" or node.children[
                        1].symbol.text == u"|=" or node.children[1].symbol.text == u"^=" or node.children[
                        1].symbol.text == u"<<=" or node.children[1].symbol.text == u">>=" or node.children[
                        1].symbol.text == u">>>="):
                assert False

            originalText = node.children[1].symbol.text

            if originalText == u"+=":
                node.children[1].symbol.oldText = u"+="
                node.children[1].symbol.text = u"-="
            elif originalText == u"-=":
                node.children[1].symbol.oldText = u"-="
                node.children[1].symbol.text = u"+="
            elif originalText == u"/=":
                node.children[1].symbol.oldText = u"/="
                node.children[1].symbol.text = u"*="
            elif originalText == u"*=":
                node.children[1].symbol.oldText = u"*="
                node.children[1].symbol.text = u"/="
            elif originalText == u"%=":
                node.children[1].symbol.oldText = u"%="
                node.children[1].symbol.text = u"/="
            elif originalText == u"&=":
                node.children[1].symbol.oldText = u"&="
                node.children[1].symbol.text = u"|="
            elif originalText == u"|=":
                node.children[1].symbol.oldText = u"|="
                node.children[1].symbol.text = u"^="
            elif originalText == u"^=":
                node.children[1].symbol.oldText = u"^="
                node.children[1].symbol.text = u"&="
            elif originalText == u">>=":
                node.children[1].symbol.oldText = u">>="
                node.children[1].symbol.text = u">>>="
            elif originalText == u"<<=":
                node.children[1].symbol.oldText = u"<<="
                node.children[1].symbol.text = u">>="
            elif originalText == u">>>=":
                node.children[1].symbol.oldText = u">>>="
                node.children[1].symbol.text = u">>="
            else:
                assert False

            return tree
